from funcy import raiser, rcompose
from .common import decorate, tap
from ..nodes import Media
from datetime import datetime



@decorate(accepts=Media, returns=Media)
def like(bot, nodes,  args):
    max = float(args['max']) if 'max' in args else float('inf')

    count = 0
    events = []

    def increment():
        nonlocal count
        count += 1
        return True

    def add_event(node: Media):
        events.append({
            'type': 'like',
            'metadata': bot.metadata,
            'args': {},
            'node': {
                'type': 'media',
                'url': node.url,
            },
            'timestamp': int(datetime.utcnow().timestamp())
        })
        return node

    stop = raiser(StopIteration)

    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        # lambda node: node \
        #     if bot.suitable(node) \
        #     else tap(None,lambda: bot.logger.warning('{} not suitable'.format(node))),
        lambda node: like_media(node, bot=bot) \
            if node else None,
        lambda x: x and increment() and x,
        lambda x: x and add_event(x),
    )


    liked = map(process, nodes)
    liked = filter(lambda x: x, liked)

    return liked, { 'events': events }


def like_media(media, bot):
        bot.api.post_like(media.pk)
        bot.logger.info(f'liked media {media}')
        bot.total['likes'] += 1
        bot.sleep('like')
        return media



import warnings

from .common import ClientDeprecationWarning
from ..compatpatch import ClientCompatPatch
from ..compat import compat_urllib_parse
from ..utils import raise_if_invalid_rank_token


class FeedEndpointsMixin(object):
    """For endpoints in ``/feed/``."""

    def feed_liked(self, **kwargs):
        """
        Get liked feed

        :param kwargs:
            - **max_id**: For pagination. Taken from ``next_max_id`` in the previous page.

        :return:
        """
        res = self._call_api('feed/liked/', query=kwargs)
        if self.auto_patch and res.get('items'):
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def feed_timeline(self, **kwargs):
        """
        Get timeline feed. To get a new timeline feed, you can mark a set of media
        as seen by setting seen_posts = comma-separated list of media IDs. Example:
        ``api.feed_timeline(seen_posts='123456789_12345,987654321_54321')``
        """
        params = {
            '_uuid': self.uuid,
            '_csrftoken': self.csrftoken,
            'is_prefetch': '0',
            'is_pull_to_refresh': '0',
            'phone_id': self.phone_id,
            'timezone_offset': self.timezone_offset,
        }
        params.update(kwargs)
        res = self._call_api('feed/timeline/', params=params, unsigned=True)
        if self.auto_patch:
            [ClientCompatPatch.media(m['media_or_ad'], drop_incompat_keys=self.drop_incompat_keys)
             if m.get('media_or_ad') else m
             for m in res.get('feed_items', [])]
        return res

    def feed_popular(self, **kwargs):   # pragma: no cover
        """Get popular feed. This endpoint is believed to be obsolete. Do not use."""
        warnings.warn(
            'This endpoint is believed to be obsolete. Do not use.',
            ClientDeprecationWarning)

        query = {
            'people_teaser_supported': '1',
            'rank_token': self.rank_token,
            'ranked_content': 'true'
        }
        query.update(kwargs)
        res = self._call_api('feed/popular/', query=query)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def user_feed(self, user_id, **kwargs):
        """
        Get the feed for the specified user id

        :param user_id:
        :param kwargs:
            - **max_id**: For pagination
            - **min_timestamp**: For pagination
        :return:
        """
        endpoint = 'feed/user/{user_id!s}/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint, query=kwargs)

        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def self_feed(self, **kwargs):
        """Get authenticated user's own feed"""
        return self.user_feed(self.authenticated_user_id, **kwargs)

    def username_feed(self, user_name, **kwargs):
        """
        Get the feed for the specified user name

        :param user_name:
        :param kwargs:
            - **max_id**: For pagination
            - **min_timestamp**: For pagination
        :return:
        """
        endpoint = 'feed/user/{user_name!s}/username/'.format(**{'user_name': user_name})
        res = self._call_api(endpoint, query=kwargs)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def reels_tray(self, **kwargs):
        """Get story reels tray"""
        res = self._call_api('feed/reels_tray/', query=kwargs)
        if self.auto_patch:
            for u in res.get('tray', []):
                if not u.get('items'):
                    continue
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in u.get('items', [])]
        return res

    def user_reel_media(self, user_id, **kwargs):
        """
        Get user story/reel media

        :param user_id:
        :param kwargs:
        :return:
        """
        endpoint = 'feed/user/{user_id!s}/reel_media/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint, query=kwargs)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def reels_media(self, user_ids, **kwargs):
        """
        Get multiple users' reel/story media

        :param user_ids: list of user IDs
        :param kwargs:
        :return:
        """
        user_ids = [str(x) for x in user_ids]
        params = {'user_ids': user_ids}
        params.update(kwargs)

        res = self._call_api('feed/reels_media/', params=params)
        if self.auto_patch:
            for reel_media in res.get('reels_media', []):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in reel_media.get('items', [])]
            for _, reel in list(res.get('reels', {}).items()):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in reel.get('items', [])]
        return res

    def feed_tag(self, tag, rank_token, **kwargs):
        """
        Get tag feed

        :param tag:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single tag feed.
        :param kwargs:
        :return:
        """
        raise_if_invalid_rank_token(rank_token)

        query_params = {
            'rank_token': rank_token
        }
        query_params.update(kwargs)
        endpoint = 'feed/tag/{tag!s}/'.format(
            **{'tag': compat_urllib_parse.quote(tag.encode('utf8'))})
        res = self._call_api(endpoint, query=query_params)
        if self.auto_patch:
            if res.get('items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('items', [])]
            if res.get('ranked_items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('ranked_items', [])]
            if res.get('story', {}).get('items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('story', {}).get('items', [])]
        return res

    def user_story_feed(self, user_id):
        """
        Get a user's story feed and current/replay broadcasts (if available)

        :param user_id:
        :return:
        """
        endpoint = 'feed/user/{user_id!s}/story/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint)
        if self.auto_patch and res.get('reel'):
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('reel', {}).get('items', [])]
        return res

    def feed_location(self, location_id, rank_token, **kwargs):
        """
        This endpoint is believed to be obsolete. Do not use. Replaced by :meth:`location_section`.

        :param location_id:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single location.
        :return:
        """
        warnings.warn(
            'This endpoint is believed to be obsolete. Do not use.',
            ClientDeprecationWarning)

        raise_if_invalid_rank_token(rank_token)

        endpoint = 'feed/location/{location_id!s}/'.format(**{'location_id': location_id})
        query_params = {
            'rank_token': rank_token,
        }
        query_params.update(kwargs)
        res = self._call_api(endpoint, query=query_params)
        if self.auto_patch:
            if res.get('items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('items', [])]
            if res.get('ranked_items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('ranked_items', [])]
            if res.get('story', {}).get('items'):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in res.get('story', {}).get('items', [])]
        return res

    def saved_feed(self, **kwargs):
        """
        Get saved photo feed

        :param kwargs:
            - **count**: Limit the number of items returned
        :return:
        """
        res = self._call_api('feed/saved/', query=kwargs)
        if self.auto_patch:
            [ClientCompatPatch.media(m['media'], drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', []) if m.get('media')]
        return res

    def feed_only_me(self, **kwargs):
        """
        Get feed of archived media

        :param kwargs
        """
        res = self._call_api('feed/only_me_feed/', query=kwargs)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res



from .node import Node
from .user import User
from funcy import fallback, silent, compose
from .schemas import story_schema
from modeller import Model
from .common import get_image_url, get_manifest, get_video_url
import traceback




# def fallback(*args):
#     first = lambda arr: arr[1:] if len(arr) > 0 else lambda: None
#     rest = lambda arr: arr[:1]if len(arr) > 0 else []
#     return  first(args)() or fallback(rest(args))



class Story(Model, Node):
    _schema = story_schema

    def _on_init(self):
        try:
            self._validate()
        except Exception as e:
            print('ERROR in validation for Story:')
            print()
            print(str(e))
            print()
            print(self._yaml())
            print()


    mpd = property(lambda self: get_manifest(self))

    image = property(lambda self: get_image_url(self))

    image = property(lambda self: get_video_url(self))

    __repr__ = lambda self: f'Story(pk={self.pk})'

    location = property(lambda self: fallback(
        lambda: self['story_locations'][0]['location'],
        lambda: self['story_locations']['location'],
        lambda: None
    ))

    geotag: compose(property, silent)(lambda self: self['story_locations'][0]['location'])

    swipeup_url = property(silent(lambda self: self['story_cta'][0]['links'][0]['webUri']))

    spotyfy_song = property(lambda self: self['story_app_attribution']['content_url'])





"""
{
    id
    latest_reel_media
    expiring_at
    seen
    can_reply
    can_reshare
    reel_type
    user {
        pk
        username
        full_name
        is_private
        profile_pic_url
        profile_pic_id
        friendship_status {
            following
            followed_by
            blocking
            muting
            is_private
            incoming_request
            outgoing_request
            is_bestie
        }
        is_verified
    }
    items[9] {
        taken_at
        pk
        id
        device_timestamp
        media_type
        code
        client_cache_key
        filter_type
        image_versions2 {
            candidates[2] {
                width
                height
                url
            }
        }
        original_width
        original_height
        caption_position
        is_reel_media
        is_dash_eligible
        video_dash_manifest
        video_codec
        number_of_qualities
        video_versions[3] {
            type
            width
            height
            url
            id
        }
        has_audio
        video_duration
        user {
            pk
            username
            full_name
            is_private
            profile_pic_url
            profile_pic_id
            is_verified
            has_anonymous_profile_picture
            reel_auto_archive
            is_unpublished
            is_favorite
        }
        caption {
            _
        }
        caption_is_edited
        photo_of_you
        can_viewer_save
        organic_tracking_token
        expiring_at
        imported_taken_at
        can_reshare
        story_hashtags[1] {
            x
            y
            z
            width
            height
            rotation
            is_pinned
            is_hidden
            hashtag {
                name
                id
            }
        }
        supports_reel_reactions
        show_one_tap_fb_share_tooltip
        has_shared_to_fb
    }
    prefetch_count
    has_besties_media
    media_count
    status
}
"""



from .common import decorate
from ..nodes import Media
from ..bot import Bot



@decorate(accepts=Media, returns=Media)
def delete(bot: Bot, nodes,  args):


    for node in nodes:
        res = bot.api.delete_media(node.id)

        if not res.get('did_delete'):
            bot.logger.warning(f'can\'t delete media {node}')
        else:
            bot.logger.info(f'deleted {node}')

        bot.sleep('delete')


    return [], {}



import warnings

from .common import ClientDeprecationWarning
from ..constants import Constants
from ..compatpatch import ClientCompatPatch


class MiscEndpointsMixin(object):
    """For miscellaneous functions."""

    def sync(self, prelogin=False):
        """Synchronise experiments."""
        if prelogin:
            params = {
                'id': self.generate_uuid(),
                'experiments': Constants.LOGIN_EXPERIMENTS
            }
        else:
            params = {
                'id': self.authenticated_user_id,
                'experiments': Constants.EXPERIMENTS
            }
            params.update(self.authenticated_params)
        return self._call_api('qe/sync/', params=params)

    def expose(self, experiment='ig_android_profile_contextual_feed'):  # pragma: no cover
        warnings.warn(
            'This endpoint is believed to be obsolete. Do not use.',
            ClientDeprecationWarning)

        params = {
            'id': self.authenticated_user_id,
            'experiment': experiment
        }
        params.update(self.authenticated_params)
        return self._call_api('qe/expose/', params=params)

    def megaphone_log(self, log_type='feed_aysf', action='seen', reason='', **kwargs):
        """
        A tracking endpoint of sorts

        :param log_type:
        :param action:
        :param reason:
        :param kwargs:
        :return:
        """
        params = {
            'type': log_type,
            'action': action,
            'reason': reason,
            '_uuid': self.uuid,
            'device_id': self.device_id,
            '_csrftoken': self.csrftoken,
            'uuid': self.generate_uuid(return_hex=True)
        }
        params.update(kwargs)
        return self._call_api('megaphone/log/', params=params, unsigned=True)

    def ranked_recipients(self):
        """Get ranked recipients"""
        res = self._call_api('direct_v2/ranked_recipients/', query={'show_threads': 'true'})
        return res

    def recent_recipients(self):
        """Get recent recipients"""
        res = self._call_api('direct_share/recent_recipients/')
        return res

    def news(self, **kwargs):
        """
        Get news feed of accounts the logged in account is following.
        This returns the items in the 'Following' tab.
        """
        return self._call_api('news/', query=kwargs)

    def news_inbox(self):
        """
        Get inbox feed of activity related to the logged in account.
        This returns the items in the 'You' tab.
        """
        return self._call_api(
            'news/inbox/', query={'limited_activity': 'true', 'show_su': 'true'})

    def direct_v2_inbox(self):
        """Get v2 inbox"""
        return self._call_api('direct_v2/inbox/')

    def oembed(self, url, **kwargs):
        """
        Get oembed info

        :param url:
        :param kwargs:
        :return:
        """
        query = {'url': url}
        query.update(kwargs)
        res = self._call_api('oembed/', query=query)
        return res

    def translate(self, object_id, object_type):
        """

        :param object_id: id value for the object
        :param object_type: One of [1, 2, 3] where
                            1 = CAPTION - unsupported
                            2 = COMMENT - unsupported
                            3 = BIOGRAPHY
        :return:
        """
        warnings.warn('This endpoint is not tested fully.', UserWarning)
        res = self._call_api(
            'language/translate/',
            query={'id': object_id, 'type': object_type})
        return res

    def bulk_translate(self, comment_ids):
        """
        Get translations of comments

        :param comment_ids: list of comment/caption IDs
        :return:
        """
        if isinstance(comment_ids, str):
            comment_ids = [comment_ids]
        query = {'comment_ids': ','.join(comment_ids)}
        res = self._call_api('language/bulk_translate/', query=query)
        return res

    def top_search(self, query):
        """
        Search for top matching hashtags, users, locations

        :param query: search terms
        :return:
        """
        res = self._call_api(
            'fbsearch/topsearch/',
            query={'context': 'blended', 'ranked_token': self.rank_token, 'query': query})
        if self.auto_patch and res.get('users', []):
            [ClientCompatPatch.list_user(u['user']) for u in res['users']]
        return res

    def stickers(self, sticker_type='static_stickers', location=None):
        """
        Get sticker assets

        :param sticker_type: One of ['static_stickers']
        :param location: dict containing 'lat', 'lng', 'horizontalAccuracy'.
                         Example: {'lat': '', 'lng': '', 'horizontalAccuracy': ''}
                         'horizontalAccuracy' is a float in meters representing the estimated horizontal accuracy
                         https://developer.android.com/reference/android/location/Location.html#getAccuracy()
        :return:
        """
        if sticker_type not in ['static_stickers']:
            raise ValueError('Invalid sticker_type: {0!s}'.format(sticker_type))
        if location and not ('lat' in location and 'lng' in location and 'horizontalAccuracy' in location):
            raise ValueError('Invalid location')
        params = {
            'type': sticker_type
        }
        if location:
            params['lat'] = location['lat']
            params['lng'] = location['lng']
            params['horizontalAccuracy'] = location['horizontalAccuracy']
        params.update(self.authenticated_params)
        return self._call_api('creatives/assets/', params=params)



from .common import decorate
from ..nodes import Node, node_classes
import json
from .common import dotdict
from ..bot import Bot
from funcy import mapcat
import time
from datetime import datetime



@decorate(accepts=(*node_classes.values(),), returns=Node)
def sleep(bot: Bot, nodes,  args):
    try:
        amount = int(args)
    except Exception:
        bot.logger.error('wait argument must be a number (of seconds)')
    time.sleep(amount)

    events = [{
        'type': 'sleep',
        'metadata': bot.metadata,
        'node': {},
        'timestamp': int(datetime.utcnow().timestamp()),
    }]
    return nodes, { 'events': events }


from ..utils import gen_user_breadcrumb
from ..compatpatch import ClientCompatPatch


class LiveEndpointsMixin(object):
    """For endpoints in ``/live/``."""

    def user_broadcast(self, user_id):
        """
        Helper method to get a user's broadcast if there is one currently live. Returns none otherwise.

        :param user_id:
        :return:
        """
        results = self.user_story_feed(user_id)
        return results.get('broadcast')

    def broadcast_like(self, broadcast_id, like_count=1):
        """
        Like a live broadcast

        :param broadcast_id: Broadcast id
        :param like_count:
        :return:
        """
        if not 1 <= like_count <= 5:
            raise ValueError('Invalid like_count')
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/like/'.format(**{'broadcast_id': broadcast_id})
        params = {'user_like_count': str(like_count)}
        params.update(self.authenticated_params)
        return self._call_api(endpoint, params=params)

    def broadcast_like_count(self, broadcast_id, like_ts=0):
        """
        Get a live broadcast's like count

        :param broadcast_id: Broadcast id
        :return:
        """
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/get_like_count/'.format(**{'broadcast_id': broadcast_id})
        return self._call_api(endpoint, query={'like_ts': like_ts})

    def broadcast_comments(self, broadcast_id, last_comment_ts=0):
        """
        Get a live broadcast's latest comments

        :param broadcast_id: Broadcast id
        :param last_comment_ts:
        :return:
        """
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/get_comment/'.format(**{'broadcast_id': broadcast_id})
        res = self._call_api(endpoint, query={'last_comment_ts': last_comment_ts})
        if self.auto_patch and res.get('comments'):
            [ClientCompatPatch.comment(c) for c in res.get('comments', [])]
            if res.get('pinned_comment'):
                ClientCompatPatch.comment(res['pinned_comment'])
        return res

    def broadcast_heartbeat_and_viewercount(self, broadcast_id):
        """
        Get a live broadcast's heartbeat and viewer count

        :param broadcast_id: Broadcast id
        :return:
        """
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/heartbeat_and_get_viewer_count/'.format(**{'broadcast_id': broadcast_id})
        params = {
            '_csrftoken': self.csrftoken,
            '_uuid': self.uuid
        }
        return self._call_api(endpoint, params=params, unsigned=True)

    def broadcast_comment(self, broadcast_id, comment_text):
        """
        Post a comment to a live broadcast

        :param broadcast_id: Broadcast id
        :param comment_text: Comment text
        :return:
        """
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/comment/'.format(**{'broadcast_id': broadcast_id})
        params = {
            'live_or_vod': '1',
            'offset_to_video_start': '0',
            'comment_text': comment_text,
            'user_breadcrumb': gen_user_breadcrumb(len(comment_text)),
            'idempotence_token': self.generate_uuid(),
        }
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        if self.auto_patch and res.get('comment'):
            ClientCompatPatch.comment(res['comment'])
        return res

    def broadcast_info(self, broadcast_id):
        """
        Get broadcast information.
        Known broadcast_status values: 'active', 'interrupted', 'stopped', 'hard_stop'

        :param broadcast_id: Broadcast Id
        :return:
            .. code-block:: javascript

                {
                  "status": "ok",
                  "broadcast_status": "active",
                  "media_id": "12345678934374208_123456789",
                  "cover_frame_url": "https://scontent-hkg3-1.cdninstagram.com/something.jpg",
                  "broadcast_owner": {
                    "username": "abc",
                    "friendship_status": {
                      "incoming_request": false,
                      "followed_by": false,
                      "outgoing_request": false,
                      "following": false,
                      "blocking": false,
                      "is_private": false
                    },
                    "profile_pic_url": "http://scontent-hkg3-1.cdninstagram.com/somethingelse.jpg",
                    "profile_pic_id": "1234567850644676241_123456789",
                    "full_name": "ABC",
                    "pk": 123456789,
                    "is_verified": true,
                    "is_private": false
                  },
                  "dash_abr_playback_url": null,
                  "broadcast_message": "",
                  "published_time": 1485312576,
                  "dash_playback_url": "https://scontent-hkg3-1.cdninstagram.com/hvideo-ash1/v/dash-hd/spmething.mpd",
                  "rtmp_playback_url": "rtmp://svelivestream007.16.ash1.facebook.com:16000/live-hd/something",
                  "id": 178591123456789,
                  "viewer_count": 9000.0
                }
        """
        broadcast_id = str(broadcast_id)
        endpoint = 'live/{broadcast_id!s}/info/'.format(**{'broadcast_id': broadcast_id})
        return self._call_api(endpoint)

    def suggested_broadcasts(self, **kwargs):
        """
        Get sugggested broadcasts

        :param kwargs:
        :return:
        """
        return self._call_api('live/get_suggested_broadcasts/', query=kwargs)

    def replay_broadcast_comments(
            self, broadcast_id, starting_offset=0,
            encoding_tag='instagram_dash_remuxed'):
        """
        Get comments for a post live broadcast.

        :param broadcast_id:
        :param starting_offset:
        :param encoding_tag:
        :return:
        """
        broadcast_id = str(broadcast_id)
        query = {
            'starting_offset': starting_offset,
            'encoding_tag': encoding_tag,
        }
        endpoint = 'live/{broadcast_id!s}/get_post_live_comments/'.format(
            **{'broadcast_id': broadcast_id})
        res = self._call_api(endpoint, query=query)
        if self.auto_patch and res.get('comments'):
            [ClientCompatPatch.comment(c['comment']) for c in res.get('comments', [])
             if c.get('comment')]
        return res

    def replay_broadcast_likes(
            self, broadcast_id, starting_offset=0,
            encoding_tag='instagram_dash_remuxed'):
        """
        Get likes for a post live broadcast.

        :param broadcast_id:
        :param starting_offset:
        :param encoding_tag:
        :return:
        """
        broadcast_id = str(broadcast_id)
        query = {
            'starting_offset': starting_offset,
            'encoding_tag': encoding_tag,
        }
        endpoint = 'live/{broadcast_id!s}/get_post_live_likes/'.format(
            **{'broadcast_id': broadcast_id})
        return self._call_api(endpoint, query=query)



from .node import Node
from .schemas import user_schema
from modeller import Model
import traceback



class User(Model, Node):
    def _on_init(self):
        try:
            self._validate()
        except Exception as e:
            print('ERROR in validation for User:')
            print()
            print(str(e))
            print()
            print(self._yaml())
            print()
            
    _schema = user_schema
    __repr__ = lambda self: f'User(pk={self.pk}, username={self.username})'

    # id = property(lambda self: self['pk'])



from .node import Node
from modeller import Model
from .schemas import comment_schema
import traceback


class Comment(Model, Node):

    def _on_init(self):
        try:
            self._validate()
        except Exception as e:
            print('ERROR in validation for Comment:')
            print()
            print(str(e))
            print()
            print(self._yaml())
            print()

    _schema = comment_schema



from funcy import mapcat, rcompose
from itertools import islice
from typing import List
from ..nodes import User, Media
from .common import decorate





@decorate(accepts=Media, returns=User)
def usertags(bot, nodes,  args) -> List[User]:

    amount = args.get('amount') or 1


    def process(node,):
        tags = node._usertags
        tags = [User(**data) for data in tags]
        # print(tags)
        yield from islice(tags, amount)

    result = mapcat(process, nodes)

    return result, {}



import json
from ..compatpatch import ClientCompatPatch


class CollectionsEndpointsMixin(object):
    """For endpoints in related to collections functionality."""

    def list_collections(self):
        return self._call_api('collections/list/')

    def collection_feed(self, collection_id, **kwargs):
        """
        Get the items in a collection.

        :param collection_id: Collection ID
        :return:
        """
        endpoint = 'feed/collection/{collection_id!s}/'.format(**{'collection_id': collection_id})
        res = self._call_api(endpoint, query=kwargs)
        if self.auto_patch and res.get('items'):
            [ClientCompatPatch.media(m['media'], drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', []) if m.get('media')]
        return res

    def create_collection(self, name, added_media_ids=None):
        """
        Create a new collection.

        :param name: Name for the collection
        :param added_media_ids: list of media_ids
        :return:
            .. code-block:: javascript

                {
                  "status": "ok",
                  "collection_id": "1700000000123",
                  "cover_media": {
                    "media_type": 1,
                    "original_width": 1080,
                    "original_height": 1080,
                    "id": 1492726080000000,
                    "image_versions2": {
                      "candidates": [
                        {
                          "url": "http://scontent-xx4-1.cdninstagram.com/...123.jpg",
                          "width": 1080,
                          "height": 1080
                        },
                        ...
                      ]
                    }
                  },
                  "collection_name": "A Collection"
                }
        """
        params = {'name': name}
        if added_media_ids and isinstance(added_media_ids, str):
            added_media_ids = [added_media_ids]
        if added_media_ids:
            params['added_media_ids'] = json.dumps(added_media_ids, separators=(',', ':'))
        params.update(self.authenticated_params)
        return self._call_api('collections/create/', params=params)

    def edit_collection(self, collection_id, added_media_ids):
        """
        Add media IDs to an existing collection.

        :param collection_id: Collection ID
        :param added_media_ids: list of media IDs
        :return: Returns same object as :meth:`create_collection`
        """
        if isinstance(added_media_ids, str):
            added_media_ids = [added_media_ids]
        params = {
            'added_media_ids': json.dumps(added_media_ids, separators=(',', ':'))
        }
        params.update(self.authenticated_params)
        endpoint = 'collections/{collection_id!s}/edit/'.format(**{'collection_id': collection_id})
        return self._call_api(endpoint, params=params)

    def delete_collection(self, collection_id):
        """
        Delete a collection.

        :param collection_id: Collection ID
        :return:
            .. code-block:: javascript

                {
                  "status": "ok"
                }
        """
        params = self.authenticated_params
        endpoint = 'collections/{collection_id!s}/delete/'.format(**{'collection_id': collection_id})
        return self._call_api(endpoint, params=params)



from .methods import methods
from functools import reduce
from .nodes import node_classes
from .support import dotdict
from collections.abc import Iterable



class MalformedScript(Exception):
    pass


def assert_good_script(script):
    for action in script['actions']:
    
        if not 'name' in action:
            raise MalformedScript('missing `name` in some action')
            
        name = action['name']
        
        if 'nodes' in action and not 'from' in action:
            raise MalformedScript(f'missing `from` in action {name}')
        
        if not 'edges' in action:
            raise MalformedScript(f' missing `edges` in action {name}')
        
        # if not 'nodes' in action:
        #     raise MalformedScript(f' missing `nodes` in action {name}')
        
        
        get_name = lambda e: list(e.keys())[0] if isinstance(e, dict) else e
        
        for edge in (get_name(edge) for edge in action['edges']):
            if not edge in methods:
                raise MalformedScript(f'unknown edge {edge} in action {name}')
            
        check, problem = check_edges(action['edges'], action.get('from', 'arg'))
        
        if not check:
            raise MalformedScript(f'wrong edges chaining in action {name}, {problem}')




def check_edges(edges, from_type,):
    
    meth = {**methods, 'starting_point': dotdict(returns=node_classes[from_type])}
    
    def good_chain(before, after):
        before = before.returns
        after = after.accepts

        if not isinstance(after, Iterable):
            return issubclass(after, before,) 
            
        else:
            return any([issubclass(x, before,) for x in after])
    
    reducer = lambda edges, last: edges + [last] \
        if edges[-1] is not None \
        and  good_chain(meth[edges[-1]], meth[last]) \
        else edges + [None]
        
    
    get_name = lambda e: list(e.keys())[0] if isinstance(e, dict) else e
        
    names = [get_name(edge) for edge in edges]
    
    # print(names)
    # print()
        
    checks = reduce(reducer, names, ['starting_point'])
    
    # print(checks)
    
    if None in checks:
        index = checks.index(None) - 1
        errored_edge = names[index]
        right_type = methods[errored_edge].accepts
        wrong_type = methods[names[index - 1]].returns if index > 0 else from_type
        problem = f'`{errored_edge}` must receive nodes of type `{right_type}`, not `{wrong_type}`'
        return False, problem
        
    else:
        return True, ''





from io import BytesIO
import sys
import codecs
import mimetypes
import random
import string
from .compat import compat_cookiejar, compat_pickle


class ClientCookieJar(compat_cookiejar.CookieJar):
    """Custom CookieJar that can be pickled to/from strings
    """
    def __init__(self, cookie_string=None, policy=None):
        compat_cookiejar.CookieJar.__init__(self, policy)
        if cookie_string:
            if isinstance(cookie_string, bytes):
                self._cookies = compat_pickle.loads(cookie_string)
            else:
                self._cookies = compat_pickle.loads(cookie_string.encode('utf-8'))

    @property
    def auth_expires(self):
        for cookie in self:
            if cookie.name in ('ds_user_id', 'ds_user'):
                return cookie.expires
        return None

    @property
    def expires_earliest(self):
        """For backward compatibility"""
        return self.auth_expires

    def dump(self):
        return compat_pickle.dumps(self._cookies)


class MultipartFormDataEncoder(object):
    """
    Modified from
    http://stackoverflow.com/questions/1270518/python-standard-library-to-post-multipart-form-data-encoded-data
    """
    def __init__(self, boundary=None):
        self.boundary = boundary or \
            ''.join(random.choice(string.ascii_letters + string.digits + '_-') for _ in range(30))
        self.content_type = 'multipart/form-data; boundary={}'.format(self.boundary)

    @classmethod
    def u(cls, s):
        if sys.hexversion < 0x03000000 and isinstance(s, str):
            s = s.decode('utf-8')
        if sys.hexversion >= 0x03000000 and isinstance(s, bytes):
            s = s.decode('utf-8')
        return s

    def iter(self, fields, files):
        """
        :param fields: sequence of (name, value) elements for regular form fields
        :param files: sequence of (name, filename, contenttype, filedata) elements for data to be uploaded as files
        :return:
        """
        encoder = codecs.getencoder('utf-8')
        for (key, value) in fields:
            key = self.u(key)
            yield encoder('--{}\r\n'.format(self.boundary))
            yield encoder(self.u('Content-Disposition: form-data; name="{}"\r\n').format(key))
            yield encoder('\r\n')
            if isinstance(value, (int, float)):
                value = str(value)
            yield encoder(self.u(value))
            yield encoder('\r\n')
        for (key, filename, contenttype, fd) in files:
            key = self.u(key)
            filename = self.u(filename)
            yield encoder('--{}\r\n'.format(self.boundary))
            yield encoder(self.u('Content-Disposition: form-data; name="{}"; filename="{}"\r\n').format(key, filename))
            yield encoder('Content-Type: {}\r\n'.format(
                contenttype or mimetypes.guess_type(filename)[0] or 'application/octet-stream'))
            yield encoder('Content-Transfer-Encoding: binary\r\n')
            yield encoder('\r\n')
            yield (fd, len(fd))
            yield encoder('\r\n')
        yield encoder('--{}--\r\n'.format(self.boundary))

    def encode(self, fields, files):
        body = BytesIO()
        for chunk, _ in self.iter(fields, files):
            body.write(chunk)
        return self.content_type, body.getvalue()



from colorama import init, Fore
import os
import operator
from string import Formatter
import random
from .support import merge, dotdict
import funcy
import json
import datetime
from .errors import Stop

def get_field_value(field_name, mapping):
    try:
        def recursive_get(field_name, mapping):

                if '.' not in field_name:
                    return mapping[field_name], True
                else:
                    *attrs, = field_name.split('.')
                    return recursive_get(".".join(attrs[1:]), mapping[attrs[0]])
        return recursive_get(field_name, mapping)

    except Exception:
        # traceback.print_exc()
        return field_name, False




def str_format_map(format_string, mapping):
    f = Formatter()
    parsed = f.parse(format_string)
    output = []
    for literal_text, field_name, format_spec, conversion in parsed:
        conversion = '!' + conversion if conversion is not None else ''
        format_spec = ':' + format_spec if format_spec else ''
        if field_name is not None:
            field_value, found = get_field_value(field_name, mapping)
            if not found:
                text = '{{{}{}{}}}'.format(field_value,
                                           conversion,
                                           format_spec)
            else:
                format_string = '{{{}{}}}'.format(conversion, format_spec)
                text = format_string.format(field_value)
        output.append(literal_text + text)
        text = ''
    return ''.join(output)

def populate_object(script, data={}):
    def recursive_populate(script, data):
        variables = data #safedotdict(**data)
        if isinstance(script, dict):
            for (key, value) in script.items():
                if isinstance(value, str):
                    if '{' in value and '}' in value:
                        try:
                            new_value = str_format_map(value, variables)
                            script[key] = new_value
                        except Exception as e:
                            init(autoreset=True)
                            print(Fore.RED + 'error in {},\n{}'.format(value, e))
                else:
                    recursive_populate(script[key], data)
        else:
            return

    script_copy = dict(**script)
    recursive_populate(script_copy, data)
    return script_copy

def populate_string( yaml_string, data={}):
    """
    max one {{  }} per line!
    """
    try:
        def replace_in_line(line):
            if '{{' in line and '}}' in line and not ('#' in line and line.index('#') < line.index('{{')): 
                begin = line.index('{{')
                end = line.index('}}', begin)
                variable_name = line[begin:end].strip().replace('{{','').replace('}}','').strip()
                try:
                    value = xeval(variable_name, data)
                    return (
                        line[:begin].replace('{{','').replace('}}','') +
                        # value if isinstance(value, (int, float)) else repr(value) +
                        repr(value) +
                        line[end:].replace('}}','').replace('{{','')
                    )
                except Stop:
                    raise
                except Exception:
                    raise Exception('yaml file needs all data to be evaluated: {{{{ {} }}}}'.format(variable_name))
            else:
                return line
        def replace_multiline(string):
            result = ''
            parts = string.split('{{')
            if len(parts) > 1:
                for part in parts:
                    if not '}}' in part:
                        result += part
                    else:
                        index = part.index('}}')
                        expr = part[0:index]
                        expr = '\n'.join([line.strip() for line in expr.splitlines()])
                        # print(repr(expr))
                        # print(expr)
                        value = repr(xeval(expr, data))
                        result += value + part[index:].replace('}}', '')
                return result
            else:
                return string
            


        result = '\n'.join(map(replace_in_line, yaml_string.splitlines()))
        result = replace_multiline(result)
        return result
    except Stop:
        print('called stop() inside script')
        return ''


def locate_variable(script):
    begin = script.index('{{')
    end = script.index('}}', begin )
    return script[begin:end].replace('{{', '').strip()

def xeval(expr, data):
    try:
        return eval(expr, {
            **{name: getattr(funcy, name) for name in funcy.__all__},
            **os.environ,
            'random': random,
            'env': dotdict(**os.environ),
            'operator': operator,
            **data,
            'data':data,
            'datetime':datetime.datetime,
            'timedelta':datetime.timedelta,
            'json':json,
            'stop': funcy.raiser(Stop),
            # User=User,
            # Story=Story,
            # Media=Media,
            # Hashtag=Hashtag,
            # Geotag=Geotag
        })
    except Stop:
        raise
    except Exception as e:
        print(f'error `{e}` in xeval for expression "{expr}"')
        raise e from None


if __name__ == '__main__':
    s = """
    ciao: ok
    other: 
        key: {{
                str(
                    'sdfsdf"sdfsdf'
                )
            }}
        another:
            obj:
                ciao: 4
    another: ok ok
    """
    print(populate_string(s))



from .node import Node
from .user import User
from .hashtag import Hashtag
from .geotag import Geotag
from .arg import Arg
from .media import Media
from .story import Story
from .comment import Comment

node_classes = dict(
    node=Node,
    user=User,
    media=Media,
    story=Story,
    hashtag=Hashtag,
    geotag=Geotag,
    arg=Arg,
    comment=Comment,
)

(node_classes,)


from .make_predicate import make_predicate
from .bot import Bot
from random import random
import json


def make_bot(script, variables):
    """

    bot:
        username: sds
        password: ...

    max_per_day:
        likes:                  50
        follow:                 20
        unfollow:               10
        ...

    delay:
        like:                   10
        usual:                  2
        ...

    """

    # if not 'bot' in script:
    #     raise Exception('no bot in script')

    data = script['bot'] if 'bot' in script else {}

    def get(var):
        if var in variables:
            return variables[var]
        elif var in data:
            return data[var]
        else:
            return None

    def get_from_script(var):
        if var in variables:
            return variables[var]
        elif var in script:
            return data[var]
        else:
            return None

    
    bot = Bot(
        settings_path=get('settings_path'),
        settings=get('settings') or None,
        username=get('username') or error(Exception('username necessary')),
        password=get('password') or error(Exception('password necessary')),
        latitude=get('latitude') or 0,
        longitude=get('longitude') or 0,
        max_per_day={key: value for key, value in script['max_per_day'].items()} if 'max_per_day' in script else {},
        # filter_predicates= [make_predicate(script['filter'], bot)] if 'filter' in script else [], # TODO should not take bot as arg
        delay={key: value for key,value in script['delay'].items()} if 'delay' in script else {},
        disable_logging=script.get('disable_logging') or False,
        log_level=script.get('log_level') or 'INFO',
        script_name=get_from_script('name') or 'not named script',
        proxy=get_from_script('proxy') or None,
    )

    if not bot.latitude or not bot.longitude:
        bot.logger.warning(
            'no latitude and longitude in script, geotag searches will probably fail')

    return bot


def error(exception):
    raise exception



from .instagram_private_api import (
    Client,
    ClientCookieExpiredError,
    ClientLoginRequiredError,
    ClientError,
    ClientLoginError
)
import gzip
from io import BytesIO
from ..bot.support import deserialize_cookie_jar
from colorlog import ColoredFormatter
import logging
import os
import time
import json
from .instagram_private_api.http import ClientCookieJar


class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, prefix):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        return '[%s] %s' % (self.prefix, msg), kwargs

Client.login = lambda self: None

class API(Client):
    def __init__(self,**kwargs):

        if 'settings' in kwargs and 'cookies' in kwargs['settings']:
            cookies = kwargs['settings']['cookies']
            jar = ClientCookieJar()
            jar._cookies = deserialize_cookie_jar(cookies)
            kwargs['settings']['cookie'] = jar.dump()
            if not 'uuid' in kwargs['settings']:
                kwargs['settings']['uuid'] = self.generate_uuid(False, seed=kwargs['username'])

        super().__init__(**kwargs)

        # Setup logging
        self.logger = logging.getLogger(kwargs.get('username',''))

        # fh = HTMLFileHandler(title=self._id, file=logs_file, mode='w')
        # fh.setLevel(logging.INFO)
        # fh.setFormatter(file_formatter())
        # self.logger.addHandler(fh)
        if not len(self.logger.handlers):
            ch = logging.StreamHandler()
            ch.setLevel(get_logging_level())
            ch.setFormatter(colred_formatter())
            self.logger.addHandler(ch)

            self.logger.setLevel(logging.DEBUG)
            self.logger = LoggerAdapter(self.logger, kwargs['username'])

    def login(self):
        return

    @staticmethod
    def _read_response(response):
        """
        Extract the response body from a http response.
        :param response:
        :return:
        """
        if response.info().get('Content-Encoding') == 'gzip':
            buf = BytesIO(response.read())
            res = gzip.GzipFile(fileobj=buf).read().decode('utf8')
        else:
            res = response.read().decode('utf8')
        if 'DEBUG' in os.environ:
            res = json.loads(res)
            res = json.dumps(res, indent=4)
        return res

    def do_login(self):
        """Login."""
        self.logger.warning('logging in')

        prelogin_params = self._call_api(
            'si/fetch_headers/',
            params='',
            query={'challenge_type': 'signup', 'guid': self.generate_uuid(True)},
            return_response=True)

        if not self.csrftoken:
            raise ClientError(
                'Unable to get csrf from prelogin.',
                error_response=self._read_response(prelogin_params))

        login_params = {
            'device_id': self.device_id,
            'guid': self.uuid,
            'adid': self.ad_id,
            'phone_id': self.phone_id,
            '_csrftoken': self.csrftoken,
            'username': self.username,
            'password': self.password,
            'login_attempt_count': '0',
        }

        login_response = self._call_api(
            'accounts/login/', params=login_params, return_response=True)

        if not self.csrftoken:
            raise ClientError(
                'Unable to get csrf from login.',
                error_response=self._read_response(login_response))

        login_json = json.loads(self._read_response(login_response))

        if not login_json.get('logged_in_user', {}).get('pk'):
            raise ClientLoginError('Unable to login.')

        if self.on_login:
            on_login_callback = self.on_login
            on_login_callback(self)


    def statistics(self):
        """
        Get insights
        :param day:
        :return:
        """

        query = {
            'locale': 'en_US',
            'vc_policy': 'insights_policy',
            'surface': 'account',
        }
        params = {
            'variables': json.dumps({
                'IgInsightsGridMediaImage_SIZE': 240,
                'timezone': 'Etc/Greenwich',
                'activityTab': 'true',
                'audienceTab': 'true',
                'contentTab': 'true',
                'query_params': json.dumps({
                    'access_token': '',
                    'id': self.authenticated_user_id
                })
            }),
            'doc_id': '1926322010754880',
            'access_token': 'undefined',
            'fb_api_caller_class': 'RelayModern',
        }
        res = self._call_api('ads/graphql/', query=query, params=params, unsigned=True)
        if not res.get('data'):
            raise ClientError(str(res),)
        if len(res.get('errors', [])):
            raise ClientError('probably not business account, \n' + str(res.get('errors')[0].get('message', '')),)
        return res

    def media_insights(self, media_id):
        params = {
            'ig_sig_key_version': self.key_version,
        }
        res = self._call_api(f'insights/media_organic_insights/{media_id}/', query=params,)
        return res

    def send_direct_item(self, users, **options):
        data = {
            'client_context': self.generate_uuid(),
            'action': 'send_item'
        }

        users = [str(x) for x in users]

        text = options.get('text', '')

        if options.get('urls'):
            data['link_text'] = text
            data['link_urls'] = json.dumps(options.get('urls'))
            item_type = 'link'

        elif options.get('media_id',) and options.get('media_type'):
            data['text'] = text
            data['media_type'] = options.get('media_type', 'photo')
            data['media_id'] = options.get('media_id', '')
            item_type = 'media_share'

        elif options.get('hashtag',):
            data['text'] = text
            data['hashtag'] = options.get('hashtag', '')
            item_type = 'hashtag'

        elif options.get('profile_user_id'):
            data['text'] = text
            data['profile_user_id'] = options.get('profile_user_id')
            item_type = 'profile'

        else:
            data['text'] = text
            item_type = 'text'

        url = 'direct_v2/threads/broadcast/{}/'.format(item_type)

        data['recipient_users'] = f"[[{','.join(users)}]]"
        if options.get('thread_id'):
            data['thread_ids'] = f"[{options.get('thread_id')}]"
        data.update(self.authenticated_params)
        return self._call_api(url, params=data, unsigned=True)



    def agree_consent1(self):
        params = {
            '_csrftoken': self.csrftoken,
            '_uid': self.authenticated_user_id,
            '_uuid': self.uuid
        }
        return self._call_api('consent/existing_user_flow/', params=params, )# unsigned=True)


    def agree_consent2(self):
        params = {
            'current_screen_key': 'qp_intro',
            'updates': json.dumps({'existing_user_intro_state': '2'}),
            '_csrftoken': self.csrftoken,
            '_uid': self.authenticated_user_id,
            '_uuid': self.uuid
        }
        return self._call_api('consent/existing_user_flow/', params=params, )# unsigned=True)


    def agree_consent3(self):
        params = {
            'current_screen_key': 'tos_and_two_age_button',
            'updates': json.dumps({'age_consent_state': '2', 'tos_data_policy_consent_state': '2'}),
            '_csrftoken': self.csrftoken,
            '_uid': self.authenticated_user_id,
            '_uuid': self.uuid
        }
        return self._call_api('consent/existing_user_flow/', params=params,)# unsigned=True)

    def mark_item_as_seen(self, thread_id, item_id):
        params = {
            'use_unified_inbox': 'true',
            'action': 'mark_seen',
            'thread_id': str(thread_id),
            'item_id': str(item_id),
            '_csrftoken': self.csrftoken,
            # '_uid': self.authenticated_user_id,
            '_uuid': self.uuid,
        }
        return self._call_api(f'direct_v2/threads/{thread_id}/items/{item_id}/seen/', params=params, unsigned=True)

    def direct_v2_thread(self, thread_id, **kwargs):
        """
        Get v2 thread
        :param thread:
        :param kwargs:
            - **cursor**: For pagination
        :return:
        """
        endpoint = 'direct_v2/threads/{thread_id!s}/'.format(**{'thread_id': thread_id})
        res = self._call_api(endpoint, query=kwargs)
        return res


    def post_album(self, medias, caption='', location=None, **kwargs):
        """
        Post an album of up to 10 photos/videos.

        :param medias: an iterable list/collection of media dict objects

            .. code-block:: javascript

                medias = [
                    {"type": "image", "size": (720, 720), "data": "..."},
                    {
                        "type": "image", "size": (720, 720),
                        "usertags": [{"user_id":4292127751, "position":[0.625347,0.4384531]}],
                        "data": "..."
                    },
                    {"type": "video", "size": (720, 720), "duration": 12.4, "thumbnail": "...", "data": "..."}
                ]

        :param caption:
        :param location:
        :return:
        """
        album_upload_id = str(int(time.time() * 1000))
        children_metadata = []
        for media in medias:
            if len(children_metadata) >= 10:
                continue
            if media.get('type', '') not in ['image', 'video']:
                raise ValueError('Invalid media type: {0!s}'.format(media.get('type', '')))
            if not media.get('data'):
                raise ValueError('Data not specified.')
            if not media.get('size'):
                raise ValueError('Size not specified.')
            if media['type'] == 'video':
                if not media.get('duration'):
                    raise ValueError('Duration not specified.')
                if not media.get('thumbnail'):
                    raise ValueError('Thumbnail not specified.')
            if not self.compatible_aspect_ratio(media['size']):
                raise ValueError('Invalid media aspect ratio.')

            if media['type'] == 'video':
                metadata = self.post_video(
                    video_data=media['data'],
                    size=media['size'],
                    duration=media['duration'],
                    thumbnail_data=media['thumbnail'],
                    is_sidecar=True
                )
            else:
                metadata = self.post_photo(
                    photo_data=media['data'],
                    size=media['size'],
                    is_sidecar=True,
                )
                if media.get('usertags'):
                    usertags = media['usertags']
                    utags = {'in': [{'user_id': str(u['user_id']), 'position': u['position']} for u in usertags]}
                    metadata['usertags'] = json.dumps(utags, separators=(',', ':'))
            children_metadata.append(metadata)

        if len(children_metadata) <= 1:
            raise ValueError('Invalid number of media objects: {0:d}'.format(len(children_metadata)))

        # configure as sidecar
        endpoint = 'media/configure_sidecar/'
        params = {
            'caption': caption,
            'client_sidecar_id': album_upload_id,
            'children_metadata': children_metadata
        }
        if location:
            media_loc = self._validate_location(location)
            params['location'] = json.dumps(media_loc)
            if 'lat' in location and 'lng' in location:
                params['geotag_enabled'] = '1'
                params['exif_latitude'] = '0.0'
                params['exif_longitude'] = '0.0'
                params['posting_latitude'] = str(location['lat'])
                params['posting_longitude'] = str(location['lng'])
                params['media_latitude'] = str(location['lat'])
                params['media_latitude'] = str(location['lng'])
        disable_comments = kwargs.pop('disable_comments', False)
        if disable_comments:
            params['disable_comments'] = '1'

        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    @classmethod
    def compatible_aspect_ratio(cls, size):
        """
        Helper method to check aspect ratio for standard uploads

        :param size: tuple of (width, height)
        :return: True/False
        """
        min_ratio, max_ratio = 4.0 / 5.0, 90.0 / 47.0 # MediaRatios.standard
        width, height = size
        this_ratio = 1.0 * width / height
        return True # min_ratio <= this_ratio <= max_ratio


def get_logging_level():
        levels = dict(
            DEBUG=logging.DEBUG,
            INFO=logging.INFO,
            WARN=logging.WARN,
            ERROR=logging.ERROR,
            CRITICAL=logging.CRITICAL,
        )

        return os.environ['LOGGING_LEVEL'] \
            if 'LOGGING_LEVEL' in os.environ and os.environ['LOGGING_LEVEL'] in levels \
            else logging.DEBUG if 'DEBUG' in os.environ else 'INFO'

def colred_formatter():
    format = '%(asctime)s | %(levelname)-8s | %(message)s'
    cformat = '%(log_color)s' + format
    date_format = '%Y-%m-%d %H:%M'
    return ColoredFormatter(cformat, date_format,
                            log_colors={'DEBUG': 'reset', 'INFO': 'green',
                                        'WARNING': 'yellow', 'ERROR': 'red',
                                        'CRITICAL': 'red'})



from .nodes import Media, User, Geotag, Hashtag, Node
from .bot import Bot
from modeller import Model

def access(node, string, bot: Bot): # attr1.attr2
    obj = node
    for attr in string.split('.'):
        if isinstance(obj, Node) and not attr in obj:
            if isinstance(obj, User):
                data = bot.api.user_info(node.pk)
                data = data['user']
                node.__init__(**data)
            elif isinstance(obj, Media):
                data = bot.api.media_info(node.pk)
                data = data['items'][0]
                node.__init__(**data)
        obj = obj[attr]

    return obj


def make_predicate(script, bot): # TODO this is fucked up
    """
    filter:

        user:
            followers:      x > 50 and x < 1000
            following:      x < 500
            bio:            any(s not in ['porn', 'sex'] for s in x)
            is_private:     x and False
            is_business:    x and False
            is_verified:    x and False

        media:
            likes:          x < 200
            comments:       x < 100
            caption:        any(s not in x for s in ['porn', 'sex'])

        hashtag:
            name:           not in [porn, sex]

        geotag:
            name:           not in {{ cities }}
            lat:            x < 1000 and x > 500
            lng:            x < 1000 and x > 500


    """
    def predicate(node, **kwargs):
        res = True
        for (k, expr) in script.items():
            try:
                val = access(node, k, bot)
                if val is None:
                    bot.logger.error(f'failed access to {k}')
                    continue
                new_res = res and eval(expr, dict(x=val,))
                if res and not new_res:
                    bot.logger.info('{} not suitable for {}'.format(node, expr))
                res = new_res
            except (KeyError, AttributeError) as e:
                bot.logger.error(f'failed to check expression in filter: {e} for {k}')

        return res

    return predicate







from funcy import take
from datetime import datetime


switch = {
    'comments': { 
        12: 'got_commented',
        66: 'got_referenced',
    },
    'relationships': {
        101: 'got_followed',
    },
    # 'requests': {},
    # 'likes': {},
    # 'usertags': {},
    # 'usertags': {},
    # 'photos_of_you': {},
    # 'comment_likes': {},
}

def notification_events(bot):
    data = bot.api.news_inbox()
    events = []
    notifications = data['new_stories']
    # get number of notifications unread
    for notification_type, count in data['counts'].items():
        if notification_type in switch and count > 0:
            filtered = [n for n in notifications if n['story_type'] in switch[notification_type].keys()]
            filtered = sorted(filtered, key=lambda x: x['args']['timestamp'])
            filtered = take(count, reversed(filtered))
            

            for n in filtered:
                events.append({
                    'type': switch[notification_type][n['story_type']],
                    'metadata': bot.metadata,
                    'args': {
                        'notification': n['args']['text'],
                    },
                    'node': {},
                    'timestamp': int(n['args']['timestamp']),
                })
    return events

    # grab the notification from the data
    # create events for them



from .support import dotdict, merge
# from .api.exceptions import InstagramApiError
from .api.instagram_private_api.errors import ClientError
from .methods import methods
import traceback

class Dont_retry(Exception):
    """
    this exception doesn't cause a retry of the edge method when it is raised.
    Useful when a method is not found in the methods object,
    because it would be just a waste of time to retry.
    """

# class Reducer(Thread):
#
#     def __init__(self, state, edges, name=None):
#         super().__init__(name=state['bot'].username)
#         self.name = state['bot'].username
#         self.logger = state['bot'].logger
#         self.edges = edges
#         self.state = state
#
#
#
#     def run(self):
#         get_name = lambda: self.edges[0]['name']
#         name = ignore((KeyError, AttributeError), 'unnamed')(get_name)()
#         self.logger.debug('reducer beginning {} action'.format(name))
#         last_state = reduce(reducer, self.edges, self.state)
#         super().set_data(last_state['data'])
#         return




def reducer(state: dotdict, edge: dotdict):

    bot = state.bot

    if len(state.errors) > 1:
        # tried multiple times
        bot.logger.warning('got lots of errors: {}, bot: {}'.format(state.errors, bot))
        state.errors = []
        # send_bot_to_phone_verifier
        # change_bot_if_neccessary
        # resolve_captcha_if_necessary
        # sleep_more
        # remove_bot_if_broken

    try:

        # if not nodes:
        #     raise Dont_retry('no nodes, {}'.format(nodes))

        method = methods.get(edge.type, None)

        if not method:
            raise Dont_retry('can\'t find method {}'.format(edge.type))

        bot.logger.info(f'# EDGE {edge.type}')


        next_nodes, next_data = method(bot, state.nodes,  edge.args)
        state.data.append(next_data)

        # bot.logger.info('{} did success on {}'.format(type, nodes))
        # bot.logger.debug('{} returned {}'.format(type, next_nodes))

        # next_data = merge(data, {'__{}__'.format(type): next_data})

        # secs = bot.delay[type] if type in bot.delay else bot.delay['usual']
        # time.sleep(secs)

    except (KeyboardInterrupt, SystemExit):
        raise

    except Dont_retry as e:
        bot.logger.error('error reducing edge {}: \"{}\" {}' \
            .format(edge.type, e.__class__.__name__, e))
        return dotdict(nodes=[], bot=bot, errors=state.errors + [e], data=state.data)

    except Exception as exc:
        bot.logger.error('error reducing edge {}: \"{}\" \n {}'.format(
            edge.type,
            exc.__class__.__name__,
            traceback.format_exc()))
        bot.sleep('error')

        # errored_state = dotdict(**merge(state, dotdict(errors=state.errors + [exc])))
        # return reducer(errored_state, edge)
        return dotdict(nodes=[], bot=bot, errors=state.errors + [exc], data=state.data)

    else:
        # all is right, no exceptions
        # bot.logger.warning(state.data)
        # bot.logger.warning(next_data)
        return dotdict(nodes=next_nodes, bot=bot, errors=[], data=state.data)



from funcy import filter as _filter
from ..make_predicate import make_predicate
from ..nodes import Node, node_classes
from .common import decorate


@decorate(accepts=(*node_classes.values(),), returns=Node)
def filter(bot, nodes,  args) -> Node:

    if args:
        predicate = make_predicate(args, bot)
        nodes = _filter(predicate, nodes)
    else:
        nodes = bot.filter(nodes)


    return nodes, {}



class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__




def merge(a, b):
    result = dict(**a)

    [result.update({x: dict(**a[x], **b[x])}) for x in set(a.keys()) & set(b.keys())
        if isinstance(a[x], dict) and isinstance(b[x], dict)]

    [result.update({x: [*a[x], *b[x]]}) for x in set(a.keys()) & set(b.keys())
        if isinstance(a[x], list) and isinstance(b[x], list)]

    [result.update({x: a[x] if x in a else b[x]}) for x in set(a.keys()) ^ set(b.keys() )]

    return result

if __name__ == '__main__':
    x = merge(
        dict(
            a=dict(b=[3,4]),
            b=[5,4,6]
        ),
        dict(
            a=dict(c=[9]),
            b=[1]
        )
    )
    print(x)



from ..nodes import Geotag, Media
from .common import decorate


@decorate(accepts=Media, returns=Geotag)
def geotag(bot, nodes,  args) -> Geotag:

    process = lambda media: Geotag(**media.location)

    result = map(process, nodes)


    return result, {}



from .common import decorate
from ..nodes import Node, node_classes
import json
from .common import dotdict
from ..bot import Bot
from funcy import mapcat



@decorate(accepts=(*node_classes.values(),), returns=Node)
def scrape(bot: Bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else float('inf')
        key = args.get('key', 'data')
        model = args['model']

    except Exception:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}


    count = 0

    def increment():
        nonlocal count
        count += 1

    data = []

    def process(node):
        """
        model:
            name:      x.full_name
            id:        x.pk
            followers: x.followers_count
        """
        if isinstance(model, dict):
            insertion = dotdict()
            for name, expr in model.items():
                insertion[name] = evaluate(expr, node, bot=bot)
        elif isinstance(model, str):
            insertion = evaluate(model, node, bot=bot)
        
        # print(json.dumps(insertion, indent=4))

        data.append(insertion)
        bot.logger.info('scraped node {} '.format(node, ))

        if count <= max:
            increment()
            yield node

        else:
             return None

    nodes = mapcat(process, nodes)

    return nodes, {key: data}



def evaluate(expr, node, bot):
    try:
        return eval(expr, dict(x=node,
            # User=User,
            # Story=Story,
            # Media=Media,
            # Hashtag=Hashtag,
            # Geotag=Geotag
        ))

    except (KeyError, AttributeError) as e:
        bot.logger.error(f'error evaluating expression {expr}')
        bot.logger.error(e)
        # bot.logger.error(treaceback.format_exc())
        return None




from typing import List
from funcy import  rcompose, mapcat
from ..nodes import User, Media
from ..bot import Bot
from .common import cycled_api_call, decorate, tap




@decorate(accepts=User, returns=Media)
def user_feed(bot: Bot, nodes,  args) -> List[Media]:

    amount = args.get('amount')

    pack_media = lambda data: Media(**data)

    process = rcompose(
        lambda user: user.username,
        # lambda x: tap(x, lambda: print(x)),
        lambda name: cycled_api_call(amount, bot, bot.api.username_feed, name, 'items'),
        lambda items: map(pack_media, items),
        # lambda x: tap(x, lambda: print(x)),

    )

    result = mapcat(process, nodes)

    return result, {}



from .node import Node
from funcy import fallback
from modeller import Model
from .schemas import geotag_schema
import traceback

class Geotag(Model, Node):
    _schema = geotag_schema

    def _on_init(self):
        try:
            self._validate()
        except Exception as e:
            print('ERROR in validation for Geotag:')
            print()
            print(str(e))
            print()
            print(self._yaml())
            print()
            
    __repr__ = lambda self: f'Geotag(pk={self.id}'

    id = property(lambda self:
        self.pk or \
        self.facebook_places_id or \
        self.external_id or \
        None
    )


"""
    pk
    name
    address
    city
    short_name
    lng
    lat
    external_source
    facebook_places_id
"""

"""
    "pk": 186886085560507,
    "name": "Kartepe",
    "address": "",
    "city": "",
    "short_name": "Kartepe",
    "lng": 30.195368,
    "lat": 40.675953,
    "external_source": "facebook_places",
    "facebook_places_id": 186886085560507
"""



# flake8: noqa
from .accounts import AccountsEndpointsMixin
from .discover import DiscoverEndpointsMixin
from .feed import FeedEndpointsMixin
from .friendships import FriendshipsEndpointsMixin
from .live import LiveEndpointsMixin
from .media import MediaEndpointsMixin
from .misc import MiscEndpointsMixin
from .locations import LocationsEndpointsMixin
from .tags import TagsEndpointsMixin
from .upload import UploadEndpointsMixin
from .users import UsersEndpointsMixin
from .usertags import UsertagsEndpointsMixin
from .collections import CollectionsEndpointsMixin
from .highlights import HighlightsEndpointsMixin
from .igtv import IGTVEndpointsMixin

from .common import (
    ClientDeprecationWarning,
    ClientPendingDeprecationWarning,
    ClientExperimentalWarning,
)



import json

from ..compat import (
    compat_urllib_request, compat_urllib_error,
    compat_http_client
)
from ..errors import (
    ErrorHandler, ClientError, ClientLoginError, ClientConnectionError
)
from ..http import MultipartFormDataEncoder
from ..compatpatch import ClientCompatPatch
from socket import timeout, error as SocketError
from ssl import SSLError
try:
    ConnectionError = ConnectionError       # pylint: disable=redefined-builtin
except NameError:  # Python 2:
    class ConnectionError(Exception):
        pass


class AccountsEndpointsMixin(object):
    """For endpoints in ``/accounts/``."""

    def login(self):
        """Login."""

        prelogin_params = self._call_api(
            'si/fetch_headers/',
            params='',
            query={'challenge_type': 'signup', 'guid': self.generate_uuid(True)},
            return_response=True)

        if not self.csrftoken:
            raise ClientError(
                'Unable to get csrf from prelogin.',
                error_response=self._read_response(prelogin_params))

        login_params = {
            'device_id': self.device_id,
            'guid': self.uuid,
            'adid': self.ad_id,
            'phone_id': self.phone_id,
            '_csrftoken': self.csrftoken,
            'username': self.username,
            'password': self.password,
            'login_attempt_count': '0',
        }

        login_response = self._call_api(
            'accounts/login/', params=login_params, return_response=True)

        if not self.csrftoken:
            raise ClientError(
                'Unable to get csrf from login.',
                error_response=self._read_response(login_response))

        login_json = json.loads(self._read_response(login_response))

        if not login_json.get('logged_in_user', {}).get('pk'):
            raise ClientLoginError('Unable to login.')

        if self.on_login:
            on_login_callback = self.on_login
            on_login_callback(self)

        # # Post-login calls in client
        # self.sync()
        # self.autocomplete_user_list()
        # self.feed_timeline()
        # self.ranked_recipients()
        # self.recent_recipients()
        # self.direct_v2_inbox()
        # self.news_inbox()
        # self.explore()

    def current_user(self):
        """Get current user info"""
        params = self.authenticated_params
        res = self._call_api('accounts/current_user/', params=params, query={'edit': 'true'})
        if self.auto_patch:
            ClientCompatPatch.user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def edit_profile(self, first_name, biography, external_url, email, phone_number, gender):
        """
        Edit profile

        :param first_name:
        :param biography:
        :param external_url:
        :param email: Required.
        :param phone_number:
        :param gender: male: 1, female: 2, unspecified: 3
        :return:
        """
        if int(gender) not in [1, 2, 3]:
            raise ValueError('Invalid gender: {0:d}'.format(int(gender)))
        if not email:
            raise ValueError('Email is required.')

        params = {
            'username': self.authenticated_user_name,
            'gender': int(gender),
            'phone_number': phone_number or '',
            'first_name': first_name or '',
            'biography': biography or '',
            'external_url': external_url or '',
            'email': email,
        }
        params.update(self.authenticated_params)
        res = self._call_api('accounts/edit_profile/', params=params)
        if self.auto_patch:
            ClientCompatPatch.user(res.get('user'))
        return res

    def remove_profile_picture(self):
        """Remove profile picture"""
        res = self._call_api(
            'accounts/remove_profile_picture/', params=self.authenticated_params)
        if self.auto_patch:
            ClientCompatPatch.user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def change_profile_picture(self, photo_data):
        """
        Change profile picture

        :param photo_data: byte string of image
        :return:
        """
        endpoint = 'accounts/change_profile_picture/'
        json_params = json.dumps(self.authenticated_params)
        hash_sig = self._generate_signature(json_params)
        fields = [
            ('ig_sig_key_version', self.key_version),
            ('signed_body', hash_sig + '.' + json_params)
        ]
        files = [
            ('profile_pic', 'profile_pic', 'application/octet-stream', photo_data)
        ]

        content_type, body = MultipartFormDataEncoder().encode(fields, files)

        headers = self.default_headers
        headers['Content-Type'] = content_type
        headers['Content-Length'] = len(body)

        endpoint_url = '{0}{1}'.format(self.api_url.format(version='v1'), endpoint)
        req = compat_urllib_request.Request(endpoint_url, body, headers=headers)
        try:
            self.logger.debug('POST {0!s}'.format(endpoint_url))
            response = self.opener.open(req, timeout=self.timeout)
        except compat_urllib_error.HTTPError as e:
            error_response = self._read_response(e)
            self.logger.debug('RESPONSE: {0:d} {1!s}'.format(e.code, error_response))
            ErrorHandler.process(e, error_response)
        except (SSLError, timeout, SocketError,
                compat_urllib_error.URLError,   # URLError is base of HTTPError
                compat_http_client.HTTPException) as connection_error:
            raise ClientConnectionError('{} {}'.format(
                connection_error.__class__.__name__, str(connection_error)))

        post_response = self._read_response(response)
        self.logger.debug('RESPONSE: {0:d} {1!s}'.format(response.code, post_response))
        json_response = json.loads(post_response)

        if self.auto_patch:
            ClientCompatPatch.user(json_response['user'], drop_incompat_keys=self.drop_incompat_keys)

        return json_response

    def set_account_private(self):
        """Make account private"""
        res = self._call_api('accounts/set_private/', params=self.authenticated_params)
        if self.auto_patch:
            ClientCompatPatch.list_user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def set_account_public(self):
        """Make account public"""""
        res = self._call_api('accounts/set_public/', params=self.authenticated_params)
        if self.auto_patch:
            ClientCompatPatch.list_user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def logout(self):
        """Logout user"""
        params = {
            'phone_id': self.phone_id,
            '_csrftoken': self.csrftoken,
            'guid': self.uuid,
            'device_id': self.device_id,
            '_uuid': self.uuid
        }
        return self._call_api('accounts/logout/', params=params, unsigned=True)

    def presence_status(self):
        """Get presence status setting"""
        json_params = json.dumps({}, separators=(',', ':'))
        query = {
            'ig_sig_key_version': self.key_version,
            'signed_body': self._generate_signature(json_params) + '.' + json_params
        }
        return self._call_api('accounts/get_presence_disabled/', query=query)

    def set_presence_status(self, disabled):
        """
        Set presence status setting

        :param disabled: True if disabling, else False
        """
        params = {
            'disabled': '1' if disabled else '0'
        }
        params.update(self.authenticated_params)
        return self._call_api('accounts/set_presence_disabled/', params=params)

    def enable_presence_status(self):
        """Enable presence status setting"""
        return self.set_presence_status(False)

    def disable_presence_status(self):
        """Disable presence status setting"""
        return self.set_presence_status(True)



from .methods import methods
from .nodes import Arg, Media, node_classes
from .support import dotdict
from funcy import rcompose
from .api.instagram_private_api.utils import InstagramID
from .api.instagram_private_api import Client
from functools import reduce
from .handle_errors import handle


def nodes_edges(body, bot):

    nodes = []
    edges = []

    if 'nodes' in body:
        nodes += body['nodes']

    if 'edges' in body:
        for edge in body['edges']:

            if isinstance(edge, dict):
                type = list(edge.keys())[0]
                args = list(edge.values())[0]
                # args = args if isinstance(args, dict) else {} # support sleep: Int
                edges += [dotdict(type=type, args=args)]
            else:
                edges += [dotdict(type=edge, args={})]
    else:
        raise Exception('in every action there must be edges')

    info = dotdict(
        name=body['name'] if 'name' in body else 'not named',
        total_nodes=reduce(calc_total_nodes, edges, len(nodes)),
    )

    edges += [dotdict(type='evaluate', args=dict(info=info))]

    nodes = initialize_nodes(nodes, body.get('from', 'arg'), bot)

    return nodes, edges



def calc_total_nodes(acc, edge):
    amount  = int(edge.args.get('amount', 1)) if isinstance(edge.args, dict) else 1
    max = edge.args.get('max', acc) if isinstance(edge.rgs, dict) else acc
    return amount * acc if acc <= max else max


def initialize_nodes(nodes, from_type, bot):

    api: Client = bot.api

    Class = node_classes[from_type.lower()]

    switch = {
        'User': lambda username: api.username_info(username)['user'],
        'Media': rcompose(
            lambda url: [x for x in url.split('/') if x][-1],
            lambda short: InstagramID.expand_code(short),
            lambda id: api.media_info(id),
            lambda data: data['items'][0],
            # lambda x: print(x) or x
        ),
        'Hashtag': lambda name: {'name': name},
        'Arg': lambda v: {'value': v},
        'Geotag': lambda name: api.location_search(bot.latitude, bot.longitude, query=name,)['venues'][0]
    }

    try:
        return handle(lambda: [Class(**switch[Class.__name__](value)) for value in nodes], bot)

    except Exception as e:
        bot.logger.error(f'error initializing nodes {nodes}: {e}')
        raise e from None
        return []



import traceback
from types import FunctionType
from xml.etree import ElementTree
from funcy import fallback

# returns the attributes of the class,
# TODO, in pyton 3.2 attribtes aren't ordered
def attributes(instance):
    list = [
        v
        for k, v in instance.__dict__.items()
        if not k.startswith('__') and not k.endswith('__')
        and not isinstance(v, (FunctionType, classmethod, staticmethod))
        ]
    return tuple(list)


class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}'



get_image_url = lambda data: fallback(
    lambda: data['image_versions2']['candidates'][0]['url'],
    lambda: data['image_versions2'][0]['url'],
    lambda: None,
)

get_video_url = lambda data: fallback(
    lambda: data['video_versions'][0]['url'],
    # lambda: get_mdp_url(data['video_dash_manifest']),
    lambda: None,
)

# get_media_url = lambda data:  get_image_url(data)


get_manifest = lambda data: data['video_dash_manifest']



import json

from ..compat import compat_urllib_parse
from ..utils import raise_if_invalid_rank_token
from ..compatpatch import ClientCompatPatch


class TagsEndpointsMixin(object):
    """For endpoints in ``/tags/``."""

    def tag_info(self, tag):
        """
        Get tag info

        :param tag:
        :return:
        """
        endpoint = 'tags/{tag!s}/info/'.format(
            **{'tag': compat_urllib_parse.quote(tag.encode('utf8'))})
        res = self._call_api(endpoint)
        return res

    def tag_related(self, tag, **kwargs):
        """
        Get related tags

        :param tag:
        :return:
        """
        endpoint = 'tags/{tag!s}/related/'.format(
            **{'tag': compat_urllib_parse.quote(tag.encode('utf8'))})
        query = {
            'visited': json.dumps([{'id': tag, 'type': 'hashtag'}], separators=(',', ':')),
            'related_types': json.dumps(['hashtag', 'location'], separators=(',', ':'))}
        res = self._call_api(endpoint, query=query)
        return res

    def tag_search(self, text, rank_token, exclude_list=[], **kwargs):
        """
        Search tag

        :param text: Search term
        :param rank_token: Required for paging through a single feed. See examples/pagination.py
        :param exclude_list: List of numerical tag IDs to exclude
        :param kwargs:
        :return:
        """
        raise_if_invalid_rank_token(rank_token)
        if not exclude_list:
            exclude_list = []
        query = {
            'q': text,
            'timezone_offset': self.timezone_offset,
            'count': 30,
            'exclude_list': json.dumps(exclude_list, separators=(',', ':')),
            'rank_token': rank_token,
        }
        query.update(kwargs)
        res = self._call_api('tags/search/', query=query)
        return res

    def tags_user_following(self, user_id):
        """
        Get tags a user is following

        :param user_id:
        :return:
        """
        endpoint = 'users/{user_id!s}/following_tags_info/'.format(user_id=user_id)
        return self._call_api(endpoint)

    def tag_follow_suggestions(self):
        """Get suggestions for tags to follow"""
        return self._call_api('tags/suggested/')

    def tag_follow(self, tag):
        """
        Follow a tag

        :param tag:
        :return:
        """
        endpoint = 'tags/follow/{hashtag!s}/'.format(
            hashtag=compat_urllib_parse.quote(tag.encode('utf-8')))
        return self._call_api(endpoint, params=self.authenticated_params)

    def tag_unfollow(self, tag):
        """
        Unfollow a tag

        :param tag:
        :return:
        """
        endpoint = 'tags/unfollow/{hashtag!s}/'.format(
            hashtag=compat_urllib_parse.quote(tag.encode('utf-8')))
        return self._call_api(endpoint, params=self.authenticated_params)

    def tag_section(self, tag, tab='top', **kwargs):
        """
        Get a tag feed section

        :param tag: tag text (without '#')
        :param tab: One of 'top', 'recent', 'places'
        :kwargs:
            **extract**: return the array of media items only
            **page**: for pagination
            **next_media_ids**: array of media_id (int) for pagination
            **max_id**: for pagination
        :return:
        """
        valid_tabs = ['top', 'recent', 'places']
        if tab not in valid_tabs:
            raise ValueError('Invalid tab: {}'.format(tab))

        extract_media_only = kwargs.pop('extract', False)
        endpoint = 'tags/{tag!s}/sections/'.format(
            **{'tag': compat_urllib_parse.quote(tag.encode('utf8'))})

        params = {
            'supported_tabs': json.dumps(valid_tabs, separators=(',', ':')),
            'tab': tab,
            'include_persistent': True,
        }

        # explicitly set known paging parameters to avoid triggering server-side errors
        if kwargs.get('max_id'):
            params['max_id'] = kwargs.pop('max_id')
        if kwargs.get('page'):
            params['page'] = kwargs.pop('page')
        if kwargs.get('next_media_ids'):
            params['next_media_ids'] = json.dumps(kwargs.pop('next_media_ids'), separators=(',', ':'))
        kwargs.pop('max_id', None)
        kwargs.pop('page', None)
        kwargs.pop('next_media_ids', None)

        params.update(kwargs)
        results = self._call_api(endpoint, params=params, unsigned=True)
        extracted_medias = []
        if self.auto_patch:
            for s in results.get('sections', []):
                for m in s.get('layout_content', {}).get('medias', []):
                    if m.get('media'):
                        ClientCompatPatch.media(m['media'], drop_incompat_keys=self.drop_incompat_keys)
                        if extract_media_only:
                            extracted_medias.append(m['media'])
        if extract_media_only:
            return extracted_medias
        return results



from funcy import take
import random
from ..nodes import Node, node_classes
from .common import decorate


@decorate(accepts=(*node_classes.values(),), returns=Node)
def shuffle(bot, nodes,  args) -> Node:

    max = args.get('max') # None works as inf
    batch = args.get('batch', max) # None works as inf
    nodes = take(batch, nodes)
    if max and max <= len(nodes):
        nodes = random.sample(nodes, k=max)
    else:
        nodes = random.sample(nodes, k=len(nodes))


    return nodes, {}



import imagesize
from funcy import take
import fleep
import json
import urllib.request
import ffmpeg
from .common import decorate, temporary_write, download_media
from ..bot import Bot
from ..nodes import Arg, Media
import traceback
from datetime import datetime


SUPPORTED_IMAGE_EXT = ['jpg']
SUPPORTED_VIDEO_EXT = ['mov', 'mp4']


@decorate(accepts=(Arg, Media), returns=Media)
def upload_post(bot: Bot, nodes,  args):
    max = args.get('max') or 1
    caption = args.get('caption') or ''
    geotag = args.get('geotag')
    disable_comments = bool(args.get('disable_comments'))

    nodes = take(max, nodes)
    


    def binary_data(node):
        is_url = isinstance(node, Arg) and 'http' in node.value
        is_media = isinstance(node, Media)
        is_path = isinstance(node, Arg) and not 'http' in node.value
        switch = {
            is_url: lambda: download_media(node.value),
            is_path: lambda: load(node.value),
            is_media: lambda: download_media(node.images[0], bot.logger.error),
        }
        return switch[True]()


    def get_geotag_data(name):
        return bot.api.location_search(bot.latitude, bot.longitude, query=name,)['venues'][0]



    if len(nodes) == 0:
        bot.logger.error('no medias to upload')

    if max == 1: # only 1 media


        node = nodes[0]
        # print(node.carousel_media)
        # print(node.sources[0])
        kwargs = dict(
              caption=caption,
              to_reel=False,
        )

        if geotag:
              kwargs.update(dict(location=get_geotag_data(geotag),))

        if disable_comments is not None:
              kwargs.update(dict(disable_comments=disable_comments))

        data = binary_data(node)
        if not data:
            bot.logger.error('skipping everything')
            return [], {}
        with temporary_write(data) as path:
                extensions = fleep.get(data[:128]).extension
                ext = extensions[0] if len(extensions) > 0 else None
                if ext in SUPPORTED_IMAGE_EXT:
                    bot.logger.info('uploading img')
                    kwargs.update(make_photo_args(data, path))
                    res = bot.api.post_photo(**kwargs)
                    uploaded_media = Media(**res.get('media',{}))
                    # print(json.dumps(res, indent=4))

                elif ext in SUPPORTED_VIDEO_EXT:
                    bot.logger.info('uploading video')
                    kwargs.update(make_video_args(data, path))
                    res = bot.api.post_video(**kwargs)
                    uploaded_media = Media(**res.get('media',{}))

                else:
                      raise RuntimeError(f'unsupportd media type {fleep.get(data[:128]).extension} for {node}')

                bot.logger.info(f'uploaded media {uploaded_media}')
                
                events = [{
                    'type': 'upload_post',
                    'metadata': bot.metadata,
                    'args': {
                        'caption': caption,
                        'geotag': geotag,
                        'caption': caption,
                        'disable_comments': disable_comments,
                    },
                    'node': {
                        'url': uploaded_media.url,
                        'is_album': False,
                        'type': 'media',
                    },
                    'timestamp': int(datetime.utcnow().timestamp()),
        }]
                return [uploaded_media], { 'events': events }



    else: # album upload

        uploads= []

        for node in nodes:

            data = binary_data(node)
            if not data:
                bot.logger.error('skipping everything')
                return [], {}

            with temporary_write(data) as path:
                extensions = fleep.get(data[:128]).extension
                ext = extensions[0] if len(extensions) > 0 else None
                if ext in SUPPORTED_IMAGE_EXT:
                    uploads.append(make_photo_args(data, path))

                elif ext in SUPPORTED_VIDEO_EXT:
                    uploads.append(make_video_args(data, path))

                else:
                    raise RuntimeError(f'unsupportd media type {fleep.get(data[:128]).extension} for {node}')

        kwargs = dict(
            caption=caption,
            medias=uploads,
        )

        if geotag:
              kwargs.update(dict(location=get_geotag_data(geotag),))

        # if disable_comments is not None:
        #       kwargs.update(dict(disable_comments=bool(disable_comments)))

        res = bot.api.post_album(**kwargs)
        uploaded_media = Media(**res.get('media',{}))
        bot.logger.info(f'uploaded album {uploaded_media}')
        events = [{
                    'type': 'upload_post',
                    'metadata': bot.metadata,
                    'args': {
                        'caption': caption,
                        'geotag': geotag,
                        'caption': caption,
                        'disable_comments': disable_comments,
                    },
                    'node': {
                        'url': uploaded_media.url,
                        'is_album': True,
                        'type': 'media',
                    },
                    'timestamp': int(datetime.utcnow().timestamp()),
        }]
        return [uploaded_media], { 'events': events }



def load(path):
    with open(path, 'rb') as f:
        return f.read()





def get_video_size(path):
        probe = ffmpeg.probe(path)
        video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)
        width = int(video_stream['width'])
        height = int(video_stream['height'])
        return width, height

def make_thumbnail(input_data, width=100):
    import subprocess
    import ffmpeg

    args = (
        ffmpeg
            .input('pipe:')
            # ... extra processing here
            .filter('scale', width, -1)
            .output('pipe:', vframes=1)
            .get_args()
    )

    p = subprocess.Popen(['ffmpeg'] + args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    output_data = p.communicate(input=input_data)[0]
    return output_data

def get_video_duration(path):
        duration = ffmpeg.probe(path)['format']['duration']
        return float(duration)

def make_photo_args(data, path):
    return dict(
        photo_data=data,
        data=data,
        size=imagesize.get(path),
        type='image',
    )

def make_video_args(data, path):
    return dict(
        video_data=data,
        data=data,
        thumbnail_data=make_thumbnail(data),
        duration=get_video_duration(path),
        size=get_video_size(path),
        type='video',
    )




from .api.instagram_private_api.errors import (
    ClientError,
    ClientLoginRequiredError,
    ClientCookieExpiredError,
    ClientConnectionError,
    ClientThrottledError,
    ClientReqHeadersTooLargeError,
)

from .bot import Bot

def handle(func, bot: Bot):
        try:
            return func()
            
        except ClientLoginRequiredError as e:
            bot.logger.error(str(e))
            bot.relogin()
            return func()

        except ClientConnectionError as e:
            bot.logger.error(str(e))

        except (ClientReqHeadersTooLargeError, ClientThrottledError) as e:
            bot.logger.error(str(e))
            bot.sleep(5 * 60)
            return func()

        except ClientError as e:
            if 'consent_required' in str(e).lower():
                bot.logger.error('catched ' + str(e))
                bot.api.agree_consent1()
                bot.api.agree_consent2()
                bot.api.agree_consent3()
                bot.api.do_login()
                return func()
            elif 'not authorized' in str(e).lower():
                return func()
            else:
                raise e from None





def memory_summary():
    # Only import Pympler when we need it. We don't want it to
    # affect our process if we never call memory_summary.
    from pympler import summary, muppy
    mem_summary = summary.summarize(muppy.get_objects())
    rows = summary.format_(mem_summary)
    return '\n'.join(rows)



# -*- coding: utf-8 -*-
import logging
import json
import re

logger = logging.getLogger(__name__)


class ClientErrorCodes(object):
    """Holds static constant values for the http error codes returned from IG"""

    INTERNAL_SERVER_ERROR = 500
    BAD_REQUEST = 400
    NOT_FOUND = 404
    TOO_MANY_REQUESTS = 429
    REQ_HEADERS_TOO_LARGE = 431


class ClientError(Exception):
    """Generic error class, catch-all for most client issues.
    """
    def __init__(self, msg, code=None, error_response=''):
        self.code = code or 0
        self.error_response = error_response
        super(ClientError, self).__init__(msg)

    @property
    def msg(self):
        return self.args[0]


class ClientLoginError(ClientError):
    """Raised when login fails."""
    pass


class ClientLoginRequiredError(ClientError):
    """Raised when login is required."""
    pass


class ClientCookieExpiredError(ClientError):
    """Raised when cookies have expired."""
    pass


class ClientThrottledError(ClientError):
    """Raised when client detects an http 429 Too Many Requests response."""
    pass


class ClientReqHeadersTooLargeError(ClientError):
    """Raised when client detects an http 431 Request Header Fields Too Large response."""
    pass


class ClientConnectionError(ClientError):
    """Raised due to network connectivity-related issues"""
    pass


class ClientCheckpointRequiredError(ClientError):
    """Raise when IG detects suspicious activity from your account"""

    @property
    def challenge_url(self):
        try:
            error_info = json.loads(self.error_response)
            return error_info.get('challenge', {}).get('url') or error_info.get('checkpoint_url')
        except ValueError as ve:
            logger.warning('Error parsing error response: {}'.format(str(ve)))
        return None


class ClientChallengeRequiredError(ClientCheckpointRequiredError):
    """Raise when IG detects suspicious activity from your account"""


class ClientSentryBlockError(ClientError):
    """Raise when IG has flagged your account for spam or abusive behavior"""
    pass


class ErrorHandler(object):

    KNOWN_ERRORS_MAP = [
        {'patterns': ['bad_password', 'invalid_user'], 'error': ClientLoginError},
        {'patterns': ['login_required'], 'error': ClientLoginRequiredError},
        {
            'patterns': ['checkpoint_required', 'checkpoint_challenge_required', 'checkpoint_logged_out'],
            'error': ClientCheckpointRequiredError
        },
        {'patterns': ['challenge_required'], 'error': ClientChallengeRequiredError},
        {'patterns': ['sentry_block'], 'error': ClientSentryBlockError},
    ]

    @staticmethod
    def process(http_error, error_response):
        """
        Tries to process an error meaningfully

        :param http_error: an instance of compat_urllib_error.HTTPError
        :param error_response: body of the error response
        """
        error_msg = http_error.reason
        if http_error.code == ClientErrorCodes.REQ_HEADERS_TOO_LARGE:
            raise ClientReqHeadersTooLargeError(
                error_msg,
                code=http_error.code,
                error_response=error_response)

        try:
            error_obj = json.loads(error_response)
            error_message_type = str(error_obj.get('error_type', '') or error_obj.get('message', ''))
            if http_error.code == ClientErrorCodes.TOO_MANY_REQUESTS:
                raise ClientThrottledError(
                    error_obj.get('message'), code=http_error.code,
                    error_response=json.dumps(error_obj))

            for error_info in ErrorHandler.KNOWN_ERRORS_MAP:
                for p in error_info['patterns']:
                    if re.search(p, error_message_type):
                        raise error_info['error'](
                            error_message_type, code=http_error.code,
                            error_response=json.dumps(error_obj)
                        )
            if error_message_type:
                error_msg = '{0!s}: {1!s}'.format(http_error.reason, error_message_type)
            else:
                error_msg = http_error.reason
        except ValueError as ve:
            # do nothing else, prob can't parse json
            logger.warning('Error parsing error response: {}'.format(str(ve)))

        raise ClientError(error_msg, http_error.code, error_response)



from .node import Node
from modeller import Model



class Arg(Model, Node):
    _schema = {
        'properties':
            {
                'value': {'type': 'string'}
            },
        'type': 'object',
    }



from .api.instagram_private_api.errors import (
    ClientError, ClientLoginError, ClientLoginRequiredError,
    ClientCookieExpiredError, ClientThrottledError, ClientConnectionError,
    ClientCheckpointRequiredError, ClientChallengeRequiredError,
    ClientSentryBlockError, ClientReqHeadersTooLargeError,
)


class Stop(Exception):
    pass


import warnings

from .common import ClientExperimentalWarning
from ..compatpatch import ClientCompatPatch
from ..utils import raise_if_invalid_rank_token


class FriendshipsEndpointsMixin(object):
    """For endpoints in ``/friendships/``."""

    def autocomplete_user_list(self):
        """User list for autocomplete"""
        res = self._call_api(
            'friendships/autocomplete_user_list/',
            query={'followinfo': 'True', 'version': '2'})
        if self.auto_patch:
            [ClientCompatPatch.list_user(user, drop_incompat_keys=self.drop_incompat_keys)
             for user in res['users']]
        return res

    def user_following(self, user_id, rank_token, **kwargs):
        """
        Get user followings

        :param user_id:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single user following.
        :param kwargs:
            - **query**: Search within the user following
            - **max_id**: For pagination
        :return:
        """
        raise_if_invalid_rank_token(rank_token)

        endpoint = 'friendships/{user_id!s}/following/'.format(**{'user_id': user_id})
        query_params = {
            'rank_token': rank_token,
        }
        query_params.update(kwargs)
        res = self._call_api(endpoint, query=query_params)
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def user_followers(self, user_id, rank_token, **kwargs):
        """
        Get user followers

        :param user_id:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single user followers.
        :param kwargs:
            - **query**: Search within the user followers
            - **max_id**: For pagination
        :return:
        """
        raise_if_invalid_rank_token(rank_token)

        endpoint = 'friendships/{user_id!s}/followers/'.format(**{'user_id': user_id})
        query_params = {
            'rank_token': rank_token,
        }
        query_params.update(kwargs)
        res = self._call_api(endpoint, query=query_params)
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def friendships_pending(self):
        """Get pending follow requests"""
        res = self._call_api('friendships/pending/')
        if self.auto_patch and res.get('users'):
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def friendships_show(self, user_id):
        """
        Get friendship status with user id

        :param user_id:
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": false,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": false,
                    "is_private": false
                }
        """
        endpoint = 'friendships/show/{user_id!s}/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint)
        return res

    def friendships_show_many(self, user_ids):
        """
        Get friendship status with mulitple user ids

        :param user_ids: list of user ids
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "friendship_statuses": {
                        "123456789": {
                            "following": false,
                            "incoming_request": true,
                            "outgoing_request": false,
                            "is_private": false
                        }
                    }
                }
        """
        if isinstance(user_ids, str):
            user_ids = [user_ids]

        params = {
            '_uuid': self.uuid,
            '_csrftoken': self.csrftoken,
            'user_ids': ','.join(user_ids)
        }
        res = self._call_api('friendships/show_many/', params=params, unsigned=True)
        return res

    def friendships_create(self, user_id):
        """
        Follow a user

        :param user_id: User id
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "friendship_status": {
                        "incoming_request": false,
                        "followed_by": false,
                        "outgoing_request": false,
                        "following": true,
                        "blocking": false,
                        "is_private": false
                    }
                }
        """
        endpoint = 'friendships/create/{user_id!s}/'.format(**{'user_id': user_id})
        params = {'user_id': user_id, 'radio_type': self.radio_type}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def friendships_destroy(self, user_id, **kwargs):
        """
        Unfollow a user

        :param user_id: User id
        :param kwargs:
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": false,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": false,
                    "is_private": false
                }
        """
        endpoint = 'friendships/destroy/{user_id!s}/'.format(**{'user_id': user_id})
        params = {'user_id': user_id, 'radio_type': self.radio_type}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def friendships_block(self, user_id):
        """
        Block a user

        :param user_id: User id
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": false,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": true,
                    "is_private": false
                }
        """
        endpoint = 'friendships/block/{user_id!s}/'.format(**{'user_id': user_id})
        params = {'user_id': user_id}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def friendships_unblock(self, user_id):
        """
        Unblock a user

        :param user_id: User id
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": false,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": false,
                    "is_private": false
                }
        """
        endpoint = 'friendships/unblock/{user_id!s}/'.format(**{'user_id': user_id})
        params = {'user_id': user_id}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def block_friend_reel(self, user_id):
        """
        Hide your stories from a specific user

        :param user_id: User id
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": true,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": true,
                    "is_private": false
                }
        """
        endpoint = 'friendships/block_friend_reel/{user_id!s}/'.format(**{'user_id': user_id})
        params = {'source': 'main_feed'}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def unblock_friend_reel(self, user_id):
        """
        Unhide your stories from a specific user

        :param user_id: User id
        :return:
            .. code-block:: javascript

                {
                    "status": "ok",
                    "incoming_request": false,
                    "is_blocking_reel": false,
                    "followed_by": false,
                    "is_muting_reel": false,
                    "outgoing_request": false,
                    "following": false,
                    "blocking": true,
                    "is_private": false
                }
        """
        endpoint = 'friendships/unblock_friend_reel/{user_id!s}/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint, params=self.authenticated_params)
        return res

    def set_reel_block_status(self, user_ids, block_status='block'):
        """
        Unhide your stories from a specific user

        :param user_ids: list of user IDs
        :param block_status: One of 'block', 'unblock'
        :return:
            .. code-block:: javascript

                {
                    "friendship_statuses": {
                        "123456790": {
                            "following": true,
                            "is_private": false,
                            "incoming_request": false,
                            "outgoing_request": false,
                            "is_blocking_reel": true,
                            "is_muting_reel": false
                        },
                        "123456791": {
                            "following": true,
                            "is_private": false,
                            "incoming_request": false,
                            "outgoing_request": false,
                            "is_blocking_reel": true,
                            "is_muting_reel": false
                        }
                    },
                    "status": "ok"
                }
        """

        if block_status not in ['block', 'unblock']:
            raise ValueError('Invalid block_status: {0!s}'.format(block_status))
        if not isinstance(user_ids, list):
            user_ids = [user_ids]
        params = {'source': 'settings'}
        user_block_statuses = {}
        for user_id in user_ids:
            user_block_statuses[str(user_id)] = block_status
        params['user_block_statuses'] = user_block_statuses
        params.update(self.authenticated_params)
        return self._call_api('friendships/set_reel_block_status/', params=params)

    def blocked_reels(self):
        """
        Get list of users from whom you've hid your stories
        """
        warnings.warn('This endpoint is experimental. Do not use.', ClientExperimentalWarning)
        res = self._call_api('friendships/blocked_reels/', params=self.authenticated_params)
        if self.auto_patch and res.get('users'):
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def enable_post_notifications(self, user_id):
        """
        Turn on post notifications for specified user.

        :param user_id:
        :return:
        """
        res = self._call_api(
            'friendships/favorite/{user_id!s}/'.format(**{'user_id': user_id}),
            params=self.authenticated_params)
        return res

    def disable_post_notifications(self, user_id):
        """
        Turn off post notifications for specified user.

        :param user_id:
        :return:
        """
        res = self._call_api(
            'friendships/unfavorite/{user_id!s}/'.format(**{'user_id': user_id}),
            params=self.authenticated_params)
        return res

    def ignore_user(self, user_id):
        """
        Ignore a user's follow request.

        :param user_id:
        :return:
        """
        params = {'user_id': user_id, 'radio_type': self.radio_type}
        params.update(self.authenticated_params)
        res = self._call_api(
            'friendships/ignore/{user_id!s}/'.format(**{'user_id': user_id}),
            params=params)
        return res

    def remove_follower(self, user_id):
        """
        Remove a follower.

        :param user_id:
        :return:
        """
        params = {'user_id': user_id, 'radio_type': self.radio_type}
        params.update(self.authenticated_params)
        res = self._call_api(
            'friendships/remove_follower/{user_id!s}/'.format(**{'user_id': user_id}),
            params=params)
        return res



import datetime
from pathlib import Path
import urllib.request
import time
from ..api import API
from ..api.instagram_private_api import  (
    ClientError,
    ClientLoginRequiredError,
    ClientCookieExpiredError,
    ClientConnectionError,
    ClientThrottledError,
    ClientLoginError,
    ClientReqHeadersTooLargeError,
    ClientCheckpointRequiredError,
    ClientChallengeRequiredError,
    ClientSentryBlockError,
)
from .support import to_json, from_json, serialize_cookie_jar
from .settings import DELAY, TOTAL, MAX_PER_DAY
import json
import portalocker
import os
import logging


class Bot:

    id = 0

    def __init__(
                 self,
                 username,
                 password,
                 settings_path=None,
                 settings={},
                 proxy=None,
                 device=None,
                 max_per_day={},
                 latitude=0,
                 longitude=0,
                 filter_predicates=[],
                 delay={},
                 disable_logging=False,
                 script_name='not named script',
                 log_level='INFO',
                 ):
        

        # TODO this is horrible
        if settings_path and not settings:
            self.settings_file = make_file(settings_path, username + '_settings.json', initial='{}')
            with open(self.settings_file, 'r') as file_data:
                settings = json.load(file_data, )
                print('Reusing settings: {0!s}'.format(self.settings_file))
        elif settings is not None:
            self.settings_file = None
        else:
            raise Exception('neither settings or settings_path present')

        # self.id = Bot.id
        self.username = username
        
        self.password = password
        self.proxy = proxy
        self.predicates = filter_predicates or []
        self.total = TOTAL
        self.delay = {**DELAY, **delay}
        self.latitude = latitude or 0
        self.longitude = longitude or 0
        self.max_per_day = {**MAX_PER_DAY, **max_per_day}
        self.script_name = script_name

        

        self.start_time = datetime.datetime.utcnow()

        if proxy:
            print('using proxy')
            proxy_handler = urllib.request.ProxyHandler({
                'http': proxy,
                'https': proxy,
            })
        else:
            proxy_handler = None


        def on_login(api: API, ):
            api.logger.debug('called on_login')
            nonlocal settings
            settings.update({**api.settings})
            cookies = api.opener.cookie_jar._cookies
            cookies = serialize_cookie_jar(cookies)
            settings['cookies'] = cookies
            del settings['cookie']
            if self.settings_file:
                with portalocker.Lock(self.settings_file, 'w', timeout=10) as outfile:
                    json.dump(settings, outfile, default=to_json, indent=4)
                    print('SAVED: {0!s}'.format(self.settings_file))
                    outfile.flush()
                    os.fsync(outfile.fileno())
            

        try:
            self.api = API(
                username=username,
                password=password,
                on_login=on_login,
                proxy_handler=proxy_handler,
                settings=settings,
            )
            if not settings.get('cookies'):
                self.api.do_login()

        except (ClientCookieExpiredError, ClientLoginRequiredError, ClientLoginError, ClientConnectionError) as e:
            print('ClientCookieExpiredError/ClientLoginRequiredError: {0!s}'.format(e))
            self.api = API(
                username=username,
                password=password,
                proxy_handler=proxy_handler,
                settings=settings.update({'cookies': {}}) if settings else {},
                on_login=on_login
            )
            self.api.do_login()
        self.logger = self.api.logger

        self.logger.setLevel(getattr(logging, log_level))

        if disable_logging:
            self.logger.setLevel(logging.CRITICAL)
            
        if os.getenv('DEBUG'):
            self.logger.setLevel(logging.DEBUG)
        

        # self.api.login()



    def __repr__(self):
        return 'Bot(username=\'{}\')'.format(self.username)


    def relogin(self):
        self.api.do_login()


    def reached_limit(self, key):
        current_date = datetime.datetime.now()
        passed_days = (current_date.date() - self.start_time.date()).days
        if passed_days > 0:
            self._reset_counters()
        return self.max_per_day[key] - self.total[key] < 0

    @property
    def metadata(self):
        return {
            'username': self.username,
            'pk': str(self.pk),
            'proxy': self.proxy,
            'lng': float(self.longitude),
            'lat': float(self.latitude),
            'start_time': str(self.start_time),
            'script_name': self.script_name,
            # 'action_name': self.action_name,
            # 'totals': self.total,
        }

    @property
    def pk(self):
        return self.api.authenticated_user_id
    

    def filter(self, nodes):
        """
        this filters every node before an interaction,
        you can customize this bychanging the predicates.
        prediacte is a function that takes a node as argument
        and returns a boolean
        """
        for predicate in self.predicates:
            nodes = filter(predicate, nodes)

        return nodes

    def suitable(self, node, **kwargs):
        """
        same as filter but only one node, returns True if node in suitable
        """
        bool = True

        for predicate in self.predicates:
            bool = bool and predicate(
                node,
                **kwargs
            )

        return bool


    def sleep(self, type='usual'):

        if type in self.delay:
            self.logger.debug('sleeping for {} seconds'.format(self.delay[type]))
            time.sleep(self.delay[type])
        else:
            self.logger.debug('sleeping for {} seconds'.format(self.delay['usual']))
            time.sleep(self.delay['usual'])


    def _reset_counters(self):
        for k in self.total:
            self.total[k] = 0
        self.start_time = datetime.datetime.now()



def make_cookie_file( file, name):
    if not file:
        file = Path(str(Path('.') / '_cookies' / name)).resolve()
        file.parent.exists() or file.parent.mkdir()
    file = Path(file)
    file.exists() or file.touch()
    return str(file.resolve())


def make_file( file, name, initial=''):
    if not file:
        file = Path(str(Path('.') / name)).resolve()
        file.parent.exists() or file.parent.mkdir()
    file = Path(file)
    if not file.exists():
         file.touch()
         with open(str(file.resolve()), 'w') as f:
             f.write(initial)
    return str(file.resolve())




def load(path):
    with open(path) as f:
        return f.read()



# Copyright (c) 2017 https://github.com/ping
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

import logging
import hmac
import hashlib
import uuid
import json
import re
import time
import random
from datetime import datetime
import gzip
from io import BytesIO
import warnings
from socket import timeout, error as SocketError
from ssl import SSLError
from .compat import (
    compat_urllib_parse, compat_urllib_error,
    compat_urllib_request, compat_urllib_parse_urlparse,
    compat_http_client)
from .errors import (
    ErrorHandler, ClientError,
    ClientLoginRequiredError, ClientCookieExpiredError,
    ClientConnectionError
)
try:  # Python 3:
    # Not a no-op, we're adding this to the namespace so it can be imported.
    ConnectionError = ConnectionError       # pylint: disable=redefined-builtin
except NameError:  # Python 2:
    class ConnectionError(Exception):
        pass

from .constants import Constants
from .http import ClientCookieJar
from .endpoints import (
    AccountsEndpointsMixin, DiscoverEndpointsMixin, FeedEndpointsMixin,
    FriendshipsEndpointsMixin, LiveEndpointsMixin, MediaEndpointsMixin,
    MiscEndpointsMixin, LocationsEndpointsMixin, TagsEndpointsMixin,
    UsersEndpointsMixin, UploadEndpointsMixin, UsertagsEndpointsMixin,
    CollectionsEndpointsMixin, HighlightsEndpointsMixin,
    IGTVEndpointsMixin,
    ClientDeprecationWarning, ClientPendingDeprecationWarning,
    ClientExperimentalWarning
)

logger = logging.getLogger(__name__)
# Force Client deprecation warnings to always appear
warnings.simplefilter('always', ClientDeprecationWarning)
warnings.simplefilter('always', ClientPendingDeprecationWarning)
warnings.simplefilter('default', ClientExperimentalWarning)


class Client(AccountsEndpointsMixin, DiscoverEndpointsMixin, FeedEndpointsMixin,
             FriendshipsEndpointsMixin, LiveEndpointsMixin, MediaEndpointsMixin,
             MiscEndpointsMixin, LocationsEndpointsMixin, TagsEndpointsMixin,
             UsersEndpointsMixin, UploadEndpointsMixin, UsertagsEndpointsMixin,
             CollectionsEndpointsMixin, HighlightsEndpointsMixin,
             IGTVEndpointsMixin, object):
    """Main API client class for the private app api."""

    API_URL = 'https://i.instagram.com/api/{version!s}/'

    USER_AGENT = Constants.USER_AGENT
    IG_SIG_KEY = Constants.IG_SIG_KEY
    IG_CAPABILITIES = Constants.IG_CAPABILITIES
    SIG_KEY_VERSION = Constants.SIG_KEY_VERSION
    APPLICATION_ID = Constants.APPLICATION_ID

    def __init__(self, username, password, **kwargs):
        """

        :param username: Login username
        :param password: Login password
        :param kwargs: See below

        :Keyword Arguments:
            - **auto_patch**: Patch the api objects to match the public API. Default: False
            - **drop_incompat_key**: Remove api object keys that is not in the public API. Default: False
            - **timeout**: Timeout interval in seconds. Default: 15
            - **api_url**: Override the default api url base
            - **cookie**: Saved cookie string from a previous session
            - **settings**: A dict of settings from a previous session
            - **on_login**: Callback after successful login
            - **proxy**: Specify a proxy ex: 'http://127.0.0.1:8888' (ALPHA)
            - **proxy_handler**: Specify your own proxy handler
        :return:
        """
        self.username = username
        self.password = password
        self.auto_patch = kwargs.pop('auto_patch', False)
        self.drop_incompat_keys = kwargs.pop('drop_incompat_keys', False)
        self.api_url = kwargs.pop('api_url', None) or self.API_URL
        self.timeout = kwargs.pop('timeout', 15)
        self.on_login = kwargs.pop('on_login', None)
        self.logger = logger

        user_settings = kwargs.pop('settings', None) or {}
        self.uuid = (
            kwargs.pop('guid', None) or kwargs.pop('uuid', None) or
            user_settings.get('uuid') or self.generate_uuid(False))
        self.device_id = (
            kwargs.pop('device_id', None) or user_settings.get('device_id') or
            self.generate_deviceid())
        # application session ID
        self.session_id = (
            kwargs.pop('session_id', None) or user_settings.get('session_id') or
            self.generate_uuid(False))
        self.signature_key = (
            kwargs.pop('signature_key', None) or user_settings.get('signature_key') or
            self.IG_SIG_KEY)
        self.key_version = (
            kwargs.pop('key_version', None) or user_settings.get('key_version') or
            self.SIG_KEY_VERSION)
        self.ig_capabilities = (
            kwargs.pop('ig_capabilities', None) or user_settings.get('ig_capabilities') or
            self.IG_CAPABILITIES)
        self.application_id = (
            kwargs.pop('application_id', None) or user_settings.get('application_id') or
            self.APPLICATION_ID)

        # to maintain backward compat for user_agent kwarg
        custom_ua = kwargs.pop('user_agent', '') or user_settings.get('user_agent')
        if custom_ua:
            self.user_agent = custom_ua
        else:
            self.app_version = (
                kwargs.pop('app_version', None) or user_settings.get('app_version') or
                Constants.APP_VERSION)
            self.android_release = (
                kwargs.pop('android_release', None) or user_settings.get('android_release') or
                Constants.ANDROID_RELEASE)
            self.android_version = int(
                kwargs.pop('android_version', None) or user_settings.get('android_version') or
                Constants.ANDROID_VERSION)
            self.phone_manufacturer = (
                kwargs.pop('phone_manufacturer', None) or user_settings.get('phone_manufacturer') or
                Constants.PHONE_MANUFACTURER)
            self.phone_device = (
                kwargs.pop('phone_device', None) or user_settings.get('phone_device') or
                Constants.PHONE_DEVICE)
            self.phone_model = (
                kwargs.pop('phone_model', None) or user_settings.get('phone_model') or
                Constants.PHONE_MODEL)
            self.phone_dpi = (
                kwargs.pop('phone_dpi', None) or user_settings.get('phone_dpi') or
                Constants.PHONE_DPI)
            self.phone_resolution = (
                kwargs.pop('phone_resolution', None) or user_settings.get('phone_resolution') or
                Constants.PHONE_RESOLUTION)
            self.phone_chipset = (
                kwargs.pop('phone_chipset', None) or user_settings.get('phone_chipset') or
                Constants.PHONE_CHIPSET)
            self.version_code = (
                kwargs.pop('version_code', None) or user_settings.get('version_code') or
                Constants.VERSION_CODE)

        cookie_string = kwargs.pop('cookie', None) or user_settings.get('cookie')
        cookie_jar = ClientCookieJar(cookie_string=cookie_string)
        if cookie_string and cookie_jar.auth_expires and int(time.time()) >= cookie_jar.auth_expires:
            raise ClientCookieExpiredError('Cookie expired at {0!s}'.format(cookie_jar.auth_expires))
        cookie_handler = compat_urllib_request.HTTPCookieProcessor(cookie_jar)

        proxy_handler = kwargs.pop('proxy_handler', None)
        if not proxy_handler:
            proxy = kwargs.pop('proxy', None)
            if proxy:
                warnings.warn('Proxy support is alpha.', UserWarning)
                parsed_url = compat_urllib_parse_urlparse(proxy)
                if parsed_url.netloc and parsed_url.scheme:
                    proxy_address = '{0!s}://{1!s}'.format(parsed_url.scheme, parsed_url.netloc)
                    proxy_handler = compat_urllib_request.ProxyHandler({'https': proxy_address})
                else:
                    raise ValueError('Invalid proxy argument: {0!s}'.format(proxy))
        handlers = []
        if proxy_handler:
            handlers.append(proxy_handler)

        # Allow user to override custom ssl context where possible
        custom_ssl_context = kwargs.pop('custom_ssl_context', None)
        try:
            https_handler = compat_urllib_request.HTTPSHandler(context=custom_ssl_context)
        except TypeError:
            # py version < 2.7.9
            https_handler = compat_urllib_request.HTTPSHandler()

        handlers.extend([
            compat_urllib_request.HTTPHandler(),
            https_handler,
            cookie_handler])
        opener = compat_urllib_request.build_opener(*handlers)
        opener.cookie_jar = cookie_jar
        self.opener = opener

        # ad_id must be initialised after cookie_jar/opener because
        # it relies on self.authenticated_user_name
        self.ad_id = (
            kwargs.pop('ad_id', None) or user_settings.get('ad_id') or
            self.generate_adid())

        if not cookie_string:   # [TODO] There's probably a better way than to depend on cookie_string
            if not self.username or not self.password:
                raise ClientLoginRequiredError('login_required', code=400)
            self.login()

        self.logger.debug('USERAGENT: {0!s}'.format(self.user_agent))
        super(Client, self).__init__()

    @property
    def settings(self):
        """Helper property that extracts the settings that you should cache
        in addition to username and password."""
        return {
            'uuid': self.uuid,
            'device_id': self.device_id,
            'ad_id': self.ad_id,
            'session_id': self.session_id,
            'cookie': self.cookie_jar.dump(),
            'created_ts': int(time.time())
        }

    @property
    def user_agent(self):
        """Returns the useragent string that the client is currently using."""
        return Constants.USER_AGENT_FORMAT.format(**{
            'app_version': self.app_version,
            'android_version': self.android_version,
            'android_release': self.android_release,
            'brand': self.phone_manufacturer,
            'device': self.phone_device,
            'model': self.phone_model,
            'dpi': self.phone_dpi,
            'resolution': self.phone_resolution,
            'chipset': self.phone_chipset,
            'version_code': self.version_code})

    @user_agent.setter
    def user_agent(self, value):
        """Override the useragent string with your own"""
        mobj = re.search(Constants.USER_AGENT_EXPRESSION, value)
        if not mobj:
            raise ValueError('User-agent specified does not fit format required: {0!s}'.format(
                Constants.USER_AGENT_EXPRESSION))
        self.app_version = mobj.group('app_version')
        self.android_release = mobj.group('android_release')
        self.android_version = int(mobj.group('android_version'))
        self.phone_manufacturer = mobj.group('manufacturer')
        self.phone_device = mobj.group('device')
        self.phone_model = mobj.group('model')
        self.phone_dpi = mobj.group('dpi')
        self.phone_resolution = mobj.group('resolution')
        self.phone_chipset = mobj.group('chipset')
        self.version_code = mobj.group('version_code')

    @staticmethod
    def generate_useragent(**kwargs):
        """
        Helper method to generate a useragent string based on device parameters

        :param kwargs:
            - **app_version**
            - **android_version**
            - **android_release**
            - **brand**
            - **device**
            - **model**
            - **dpi**
            - **resolution**
            - **chipset**
        :return: A compatible user agent string
        """
        return Constants.USER_AGENT_FORMAT.format(**{
            'app_version': kwargs.pop('app_version', None) or Constants.APP_VERSION,
            'android_version': int(kwargs.pop('android_version', None) or Constants.ANDROID_VERSION),
            'android_release': kwargs.pop('android_release', None) or Constants.ANDROID_RELEASE,
            'brand': kwargs.pop('phone_manufacturer', None) or Constants.PHONE_MANUFACTURER,
            'device': kwargs.pop('phone_device', None) or Constants.PHONE_DEVICE,
            'model': kwargs.pop('phone_model', None) or Constants.PHONE_MODEL,
            'dpi': kwargs.pop('phone_dpi', None) or Constants.PHONE_DPI,
            'resolution': kwargs.pop('phone_resolution', None) or Constants.PHONE_RESOLUTION,
            'chipset': kwargs.pop('phone_chipset', None) or Constants.PHONE_CHIPSET,
            'version_code': kwargs.pop('version_code', None) or Constants.VERSION_CODE})

    @staticmethod
    def validate_useragent(value):
        """
        Helper method to validate a useragent string for format correctness

        :param value:
        :return:
        """
        mobj = re.search(Constants.USER_AGENT_EXPRESSION, value)
        if not mobj:
            raise ValueError(
                'User-agent specified does not fit format required: {0!s}'.format(
                    Constants.USER_AGENT_EXPRESSION))
        parse_params = {
            'app_version': mobj.group('app_version'),
            'android_version': int(mobj.group('android_version')),
            'android_release': mobj.group('android_release'),
            'brand': mobj.group('manufacturer'),
            'device': mobj.group('device'),
            'model': mobj.group('model'),
            'dpi': mobj.group('dpi'),
            'resolution': mobj.group('resolution'),
            'chipset': mobj.group('chipset'),
            'version_code': mobj.group('version_code'),
        }
        return {
            'user_agent': Constants.USER_AGENT_FORMAT.format(**parse_params),
            'parsed_params': parse_params
        }

    def get_cookie_value(self, key, domain=''):
        now = int(time.time())
        eternity = now + 100 * 365 * 24 * 60 * 60   # future date for non-expiring cookies
        if not domain:
            domain = compat_urllib_parse_urlparse(self.API_URL).netloc

        for cookie in sorted(
                self.cookie_jar, key=lambda c: c.expires or eternity, reverse=True):
            # don't return expired cookie
            if cookie.expires and cookie.expires < now:
                continue
            # cookie domain may be i.instagram.com or .instagram.com
            cookie_domain = cookie.domain
            # simple domain matching
            if cookie_domain.startswith('.'):
                cookie_domain = cookie_domain[1:]
            if not domain.endswith(cookie_domain):
                continue

            if cookie.name.lower() == key.lower():
                return cookie.value

        return None

    @property
    def csrftoken(self):
        """The client's current csrf token"""
        return self.get_cookie_value('csrftoken')

    @property
    def token(self):
        """For compatibility. Equivalent to :meth:`csrftoken`"""
        return self.csrftoken

    @property
    def authenticated_user_id(self):
        """The current authenticated user id"""
        return self.get_cookie_value('ds_user_id')

    @property
    def authenticated_user_name(self):
        """The current authenticated user name"""
        return self.get_cookie_value('ds_user')

    @property
    def phone_id(self):
        """Current phone ID. For use in certain functions."""
        return self.generate_uuid(return_hex=False, seed=self.device_id)

    @property
    def timezone_offset(self):
        """Timezone offset in seconds. For use in certain functions."""
        return int(round((datetime.now() - datetime.utcnow()).total_seconds()))

    @property
    def rank_token(self):
        if not self.authenticated_user_id:
            return None
        return '{0!s}_{1!s}'.format(self.authenticated_user_id, self.uuid)

    @property
    def authenticated_params(self):
        return {
            '_csrftoken': self.csrftoken,
            '_uuid': self.uuid,
            '_uid': self.authenticated_user_id
        }

    @property
    def cookie_jar(self):
        """The client's cookiejar instance."""
        return self.opener.cookie_jar

    @property
    def default_headers(self):
        return {
            'User-Agent': self.user_agent,
            'Connection': 'close',
            'Accept': '*/*',
            'Accept-Language': 'en-US',
            'Accept-Encoding': 'gzip, deflate',
            'X-IG-Capabilities': self.ig_capabilities,
            'X-IG-Connection-Type': 'WIFI',
            'X-IG-Connection-Speed': '{0:d}kbps'.format(random.randint(1000, 5000)),
            'X-IG-App-ID': self.application_id,
            'X-IG-Bandwidth-Speed-KBPS': '-1.000',
            'X-IG-Bandwidth-TotalBytes-B': '0',
            'X-IG-Bandwidth-TotalTime-MS': '0',
            'X-FB-HTTP-Engine': Constants.FB_HTTP_ENGINE,
        }

    @property
    def radio_type(self):
        """For use in certain endpoints"""
        return 'wifi-none'

    def _generate_signature(self, data):
        """
        Generates the signature for a data string

        :param data: content to be signed
        :return:
        """
        return hmac.new(
            self.signature_key.encode('ascii'), data.encode('ascii'),
            digestmod=hashlib.sha256).hexdigest()

    @classmethod
    def generate_uuid(cls, return_hex=False, seed=None):
        """
        Generate uuid

        :param return_hex: Return in hex format
        :param seed: Seed value to generate a consistent uuid
        :return:
        """
        if seed:
            m = hashlib.md5()
            m.update(seed.encode('utf-8'))
            new_uuid = uuid.UUID(m.hexdigest())
        else:
            new_uuid = uuid.uuid1()
        if return_hex:
            return new_uuid.hex
        return str(new_uuid)

    @classmethod
    def generate_deviceid(cls, seed=None):
        """
        Generate an android device ID

        :param seed: Seed value to generate a consistent device ID
        :return:
        """
        return 'android-{0!s}'.format(cls.generate_uuid(True, seed)[:16])

    def generate_adid(self, seed=None):
        """
        Generate an Advertising ID based on the login username since
        the Google Ad ID is a personally identifying but resettable ID.

        :return:
        """
        modified_seed = seed or self.authenticated_user_name or self.username
        if modified_seed:
            # Do some trivial mangling of original seed
            sha2 = hashlib.sha256()
            sha2.update(modified_seed.encode('utf-8'))
            modified_seed = sha2.hexdigest()
        return self.generate_uuid(False, modified_seed)

    @staticmethod
    def _read_response(response):
        """
        Extract the response body from a http response.

        :param response:
        :return:
        """
        if response.info().get('Content-Encoding') == 'gzip':
            buf = BytesIO(response.read())
            res = gzip.GzipFile(fileobj=buf).read().decode('utf8')
        else:
            res = response.read().decode('utf8')
        return res

    def _call_api(self, endpoint, params=None, query=None, return_response=False, unsigned=False, version='v1'):
        """
        Calls the private api.

        :param endpoint: endpoint path that should end with '/', example 'discover/explore/'
        :param params: POST parameters
        :param query: GET url query parameters
        :param return_response: return the response instead of the parsed json object
        :param unsigned: use post params as-is without signing
        :param version: for the versioned api base url. Default 'v1'.
        :return:
        """
        url = '{0}{1}'.format(self.api_url.format(version=version), endpoint)
        if query:
            url += ('?' if '?' not in endpoint else '&') + compat_urllib_parse.urlencode(query)

        headers = self.default_headers
        data = None
        if params or params == '':
            headers['Content-type'] = 'application/x-www-form-urlencoded; charset=UTF-8'
            if params == '':    # force post if empty string
                data = ''.encode('ascii')
            else:
                if not unsigned:
                    json_params = json.dumps(params, separators=(',', ':'))
                    hash_sig = self._generate_signature(json_params)
                    post_params = {
                        'ig_sig_key_version': self.key_version,
                        'signed_body': hash_sig + '.' + json_params
                    }
                else:
                    # direct form post
                    post_params = params
                data = compat_urllib_parse.urlencode(post_params).encode('ascii')

        req = compat_urllib_request.Request(url, data, headers=headers)
        try:
            self.logger.debug('REQUEST: {0!s} {1!s}'.format(url, req.get_method()))
            self.logger.debug('DATA: {0!s}'.format(data))
            response = self.opener.open(req, timeout=self.timeout)
        except compat_urllib_error.HTTPError as e:
            error_response = self._read_response(e)
            self.logger.debug('RESPONSE: {0:d} {1!s}'.format(e.code, json.dumps(json.loads(error_response), indent=4)))
            ErrorHandler.process(e, error_response)

        except (SSLError, timeout, SocketError,
                compat_urllib_error.URLError,   # URLError is base of HTTPError
                compat_http_client.HTTPException,
                ConnectionError) as connection_error:
            raise ClientConnectionError('{} {}'.format(
                connection_error.__class__.__name__, str(connection_error)))

        if return_response:
            return response

        response_content = self._read_response(response)
        self.logger.debug('RESPONSE: {0:d} {1!s}'.format(response.code, json.dumps(json.loads(response_content), indent=4)))
        json_response = json.loads(response_content)

        if json_response.get('message', '') == 'login_required':
            raise ClientLoginRequiredError(
                json_response.get('message'), code=response.code,
                error_response=json.dumps(json_response))

        # not from oembed or insights
        if not json_response.get('data') and json_response.get('provider_url') and json_response.get('status', '') != 'ok':
            raise ClientError(
                json_response.get('message', 'Unknown error'), code=response.code,
                error_response=json.dumps(json_response))

        return json_response



from funcy import mapcat, rcompose
from ..nodes import User, Media
from .common import decorate


@decorate(accepts=Media, returns=User)
def author(bot, nodes,  args):

    process = rcompose(
        lambda media: media['user'],
        lambda data: User(**data),
    )

    result = map(process, nodes)

    return result, {}



# flake8: noqa

from .client import Client
from .compatpatch import ClientCompatPatch
from .errors import (
    ClientError, ClientLoginError, ClientLoginRequiredError,
    ClientCookieExpiredError, ClientThrottledError, ClientConnectionError,
    ClientCheckpointRequiredError, ClientChallengeRequiredError,
    ClientSentryBlockError, ClientReqHeadersTooLargeError,
)
from .endpoints.upload import MediaRatios
from .endpoints.common import MediaTypes


__version__ = '1.6.0'




from .bot import Bot
from datetime import datetime


def direct_inbox_events(bot: Bot):
    data = bot.api.direct_v2_inbox()
    my_pk = str(data['viewer']['pk'])
    unseen_count = data['inbox']['unseen_count']
    if unseen_count > 0:
        for thread in data['inbox']['threads']:
            last_seen_at = int(thread['last_seen_at'][my_pk]['timestamp'])
            thread_id = thread['thread_id']
            if int(thread['last_activity_at']) > last_seen_at:
                res = bot.api.direct_v2_thread(thread_id)
                for item in res['thread']['items']:
                    if str(item['user_id']) != my_pk and int(item['timestamp']) > last_seen_at:
                        item_type = item['item_type']
                        item_id = item['item_id']
                        if item_type == 'text':
                            bot.api.mark_item_as_seen(thread_id, item_id)
                            pk = str(item['user_id'])
                            users = [user for user in thread['users'] if pk == str(user['pk'])]
                            username = users[0]['username'] if len(users) else ''
                            yield ({
                                'type': 'message',
                                # 'metadata': bot.metadata, TODO
                                'args': {
                                    'message': item['text'],
                                    'thread_id': thread_id,
                                    'item_id': item_id,
                                },
                                'node': {
                                    'type': 'user',
                                    'pk': pk,
                                    'username': username,
                                },
                                'timestamp': int(item['timestamp'])
                            })
                        else: # TODO
                            bot.logger.warning(f'skipping message of type {item_type} as not implemented')





from typing import List
from funcy import  rcompose, mapcat
from ..bot import Bot
from ..nodes import  Media, User
from .common import decorate




@decorate(accepts=Media, returns=User)
def likers(bot, nodes,  args) -> List[Media]:

    pack_user = lambda item: User(**item)
    amount = args.get('amount')

    process = rcompose(
            lambda media: media.pk,
            lambda id: get_likers(id, bot, amount),
            lambda gen: map(pack_user, gen)
        )

    result = mapcat(process, nodes)


    return result, {}





def get_likers(id, bot: Bot, amount):
    data = bot.api.media_likers(id)
    if 'users' in data:
        yield from data['users'][:amount]
    else:
        yield from []




from typing import List
from funcy import  rcompose, mapcat
from ..bot import Bot
from ..nodes import Story, User
from .common import decorate, tap




@decorate(accepts=User, returns=Story)
def user_stories(bot, nodes,  args) -> List[Story]:

    amount = args.get('amount') or 1
    pack_story = lambda data: Story(**data)
    # unmasked = lambda: unmask(bot.last)
    # log_unmasked = lambda: bot.logger.warning(unmasked())
    # bot.logger.warning([x for x in nodes])


    process = rcompose(
        lambda user: user.pk,
        # lambda id: tap(id, lambda: bot.api.get_user_stories(id)),
        # lambda x: tap(x, lambda: print(x)),
        lambda id: get_stories(bot, id, amount),
        # lambda gen: map(lambda data: print(Story(**data)._yaml()) or data, gen),
        lambda gen: map(pack_story, gen),

    )

    stories = mapcat(process, nodes)

    return stories, {}

def get_stories(bot: Bot, user_id, amount,):
    count = 0
    data = bot.api.user_story_feed(user_id)
    # import json
    # print(json.dumps(data['reel']['items'], indent=4))
    if data['reel']:
        yield from data['reel']['items'][:amount]
    else:
        bot.logger.debug('no stories')
        yield from []




schema = """

properties:
  broadcast:
    type: 'null'
  reel:
    properties:
      can_reply:
        type: boolean
      can_reshare:
        type: boolean
      expiring_at:
        type: integer
      has_besties_media:
        type: boolean
      id:
        type: integer
      items:
        type: array
        items:
          $ref: story.yaml
      latest_reel_media:
        type: integer
      media_count:
        type: integer
      prefetch_count:
        type: integer
      reel_type:
        type: string
      seen:
        type: integer
      user:
        $ref: user.yaml
    required:
      - can_reply
      - can_reshare
      - expiring_at
      - items
      - media_count
      - prefetch_count
      - reel_type
      - seen
      - user
    type: object
  status:
    type: string
required:
  - reel
  - status
type: object


"""



# flake8: noqa
# pylint: disable=unused-import
try:
    import urllib.request as compat_urllib_request
except ImportError:  # Python 2
    import urllib2 as compat_urllib_request

try:
    import urllib.error as compat_urllib_error
except ImportError:  # Python 2
    import urllib2 as compat_urllib_error

try:
    import urllib.parse as compat_urllib_parse
except ImportError:  # Python 2
    import urllib as compat_urllib_parse

try:
    from urllib.parse import urlparse as compat_urllib_parse_urlparse
except ImportError:  # Python 2
    from urlparse import urlparse as compat_urllib_parse_urlparse

try:
    import http.cookiejar as compat_cookiejar
except ImportError:  # Python 2
    import cookielib as compat_cookiejar

try:
    import http.cookies as compat_cookies
except ImportError:  # Python 2
    import Cookie as compat_cookies

try:
    import cPickle as compat_pickle
except ImportError:
    import pickle as compat_pickle

try:
    import http.client as compat_http_client
except ImportError:  # Python 2
    import httplib as compat_http_client



import warnings

from .common import ClientDeprecationWarning
from ..compatpatch import ClientCompatPatch


class DiscoverEndpointsMixin(object):
    """For endpoints in ``/discover/``."""

    def explore(self, **kwargs):
        """
        Get explore items

        :param kwargs:
            - **max_id**: For pagination
        :return:
        """
        query = {'is_prefetch': 'false', 'is_from_promote': 'false'}
        query.update(kwargs)
        res = self._call_api('discover/explore/', query=query)
        if self.auto_patch:
            [ClientCompatPatch.media(item['media'], drop_incompat_keys=self.drop_incompat_keys)
             if item.get('media') else item for item in res['items']]
        return res

    def discover_channels_home(self):       # pragma: no cover
        """Discover channels home"""
        warnings.warn(
            'This endpoint is believed to be obsolete. Do not use.',
            ClientDeprecationWarning)

        res = self._call_api('discover/channels_home/')
        if self.auto_patch:
            for item in res.get('items', []):
                for row_item in item.get('row_items', []):
                    if row_item.get('media'):
                        ClientCompatPatch.media(row_item['media'])
        return res

    def discover_chaining(self, user_id):
        """
        Get suggested users

        :param user_id:
        :return:
        """
        res = self._call_api('discover/chaining/', query={'target_id': user_id})
        if self.auto_patch:
            [ClientCompatPatch.list_user(user) for user in res.get('users', [])]
        return res

    def discover_top_live(self, **kwargs):
        """
        Get top live broadcasts

        :param kwargs:
            - max_id: For pagination
        :return:
        """
        return self._call_api('discover/top_live/', query=kwargs)

    def top_live_status(self, broadcast_ids):
        """
        Get status for a list of broadcast_ids

        :return:
        """
        if isinstance(broadcast_ids, str):
            broadcast_ids = [broadcast_ids]
        broadcast_ids = [str(x) for x in broadcast_ids]
        params = {'broadcast_ids': broadcast_ids}
        params.update(self.authenticated_params)
        return self._call_api('discover/top_live_status/', params=params)



from funcy import ignore, raiser, rcompose
from random import choice
from ..bot import Bot
from .common import decorate, extract_urls, substitute_vars, tap
from ..nodes import User, Node
import json
from datetime import datetime


@decorate(accepts=User, returns=Node)
def message(bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else float('inf')
        messages = args['messages']
        media = args.get('media_share')
        profile = args.get('profile')
        hashtag = args.get('hashtag')
    except Exception:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}

    count = 0
    events = []

    def increment():
        nonlocal count
        count += 1
        return True

    def add_event(pair):
        node, text = pair
        events.append({
            'type': 'message',
            'metadata': bot.metadata,
            'args': {
                'message': text,
            },
            'node': {
                'type': 'user',
                'username': node.username,
            },
            'timestamp': int(datetime.utcnow().timestamp())
        })
        return node

    stop = raiser(StopIteration)

    listmap = rcompose(map, list)

    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        lambda x: (x, listmap(choice, messages)),
        lambda pair: listmap(lambda m: send_message(bot, m, pair[0]), pair[1]),
        # lambda x: print(x) or x,
        lambda x: listmap(add_event, x),
        lambda x: x and x[0],
        lambda x: x and increment() and x,
        
    )

    result = map(process, nodes)
    result = filter(lambda x: x, result)

    return result, { 'events': events }




def send_message(bot: Bot, text, node, thread_id=None):

    evaluated_text = substitute_vars(text,
        receiver=node,
    )
    
    urls = extract_urls(text)
    

    res = bot.api.send_direct_item(
        users=[node.pk],
        text=evaluated_text,
        thread=thread_id,
        urls=urls
    )
    
    # print(json.dumps(res, indent=4))
    
    thread_id = res.get('payload',{}).get('thread_id', '')
    
    bot.logger.info('messaged %s' % node)
    bot.total['messages'] += 1
    bot.sleep('message')
    return (node, text)


# def send_messages(bot, text, user_ids):
#     broken_items = []
#     if not user_ids:
#         bot.logger.info("User must be at least one.")
#         return broken_items
#     bot.logger.info("Going to send %d messages." % (len(user_ids)))
#     for user in tqdm(user_ids):
#         if not bot.send_message(text, user):
#             bot.error_delay()
#             broken_items = user_ids[user_ids.index(user):]
#             break
#     return broken_items



from funcy import mapcat
from itertools import islice
from ..nodes import Hashtag, Media
from .common import decorate


@decorate(accepts=Media, returns=Hashtag)
def hashtags(bot, nodes,  args):

    amount = args.get('amount')

    def process(node,):
        text = node['caption']['text']
        raw_tags = set(part[1:] for part in text.split() if part.startswith('#'))
        tags = [Hashtag(name=tag) for tag in raw_tags]
        yield from islice(tags, amount)

    result = mapcat(process, nodes)
    # result =

    return result, {}



from funcy import ignore
from ..bot import Bot
from .common import decorate, download_media
from ..nodes import node_classes, User
import urllib.request


def load(path):
    with open(path, 'rb') as f:
        return f.read()

"""
Root:
    external_url: Str
    phone_number: Str
    username: Str
    first_name: Str
    biography: Str
    email: Str
    gender: Str
    privacy: "public" | "private"
    profile_pic: Path | Url

Path: Str
Url: Str
"""

@decorate(accepts=node_classes.values(), returns=User)
def edit_profile(bot: Bot, nodes,  args):

    @ignore((KeyError, AttributeError), None)
    def pick(key, default=None):
        return args.get(key, default)

    mode = pick('privacy')
    profile_pic = pick('profile_picture')
    # first_name, biography, external_url, email, phone_number, gender
    edits = {
        'external_url': pick('external_url'),
        'phone_number': pick('phone_number'),
        # 'username': pick('username'),
        'first_name': pick('first_name'),
        'biography': pick('biography'),
        'email': pick('email'),
        'gender': 2 if 'f' in pick('gender', '').lower() else 1,
    }

    if mode:
        if mode == 'public':
            bot.api.set_account_private()
        elif mode == 'private':
            bot.api.set_account_public()
        else:
            bot.logger.error('{} is not either "public" or "private"'.format(mode))

    
    if profile_pic:
        if 'http' in profile_pic:
            data = download_media(profile_pic, bot.logger.error)
            if data:
                bot.api.change_profile_picture(data)
        else:
            bot.api.change_profile_picture(load(profile_pic))

    # TODO set default values in edits
    if any([value for value in edits.values()]):
        old_values = bot.api.current_user()['user']
        new_values = {key: value if value is not None else old_values[key] for key, value in edits.items()}
        bot.api.edit_profile(**new_values)

    
    bot.logger.info('changed profile values')

    me = bot.api.current_user()['user']
    me = User(**me)

    return [me], {}




import fastjsonschema
import yaml
import collections
import os.path



def load(path):
    current = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(current, path)
    with open(path, encoding="utf-8") as f:
        return yaml.safe_load(f.read())

geotag_schema = load('geotag.yaml')
user_schema = load('user.yaml')
media_schema = load('media.yaml')
comment_schema = load('comment.yaml')
story_schema = load('story.yaml')


store = dict(
        geotag_schema=geotag_schema,
        user_schema=user_schema,
        media_schema=media_schema,
        comment_schema=comment_schema,
        story_schema=story_schema,
)


def resolve_refs(spec, uri='', store={}):

    resolver = fastjsonschema.RefResolver(uri, spec, store=store)

    def _do_resolve(node):
        if isinstance(node, collections.abc.Mapping) and '$ref' in node:
            with resolver.resolving(node['$ref']) as resolved:
                return resolved
        elif isinstance(node, collections.abc.Mapping):
            for k, v in node.items():
                node[k] = _do_resolve(v)
        elif isinstance(node, (list, tuple)):
            for i in range(len(node)):
                node[i] = _do_resolve(node[i])
        return node

    return _do_resolve(spec)

geotag_schema = resolve_refs(geotag_schema, store=store)

user_schema = resolve_refs(user_schema, store=store)

media_schema = resolve_refs(media_schema, store=store)

comment_schema = resolve_refs(comment_schema, store=store)

story_schema = resolve_refs(story_schema, store=store)



from typing import List
from ..bot import Bot
from funcy import  rcompose, mapcat
from ..nodes import  Media, Hashtag, Story
from .common import decorate, cycled_api_call



@decorate(accepts=Hashtag, returns=Media)
def hashtag_stories(bot: Bot, nodes,  args) -> List[Media]:
    amount = args.get('amount') or 1

    pack_story = lambda data: Story(**data)

    process = rcompose(
        lambda tag: tag.name,
        # lambda x: tap(x, lambda: print(bot.last)),
        lambda id: cycled_api_call(amount, bot, bot.api.feed_tag, id, ['story', 'items']),
        lambda items: map(pack_story, items),
    )

    result = mapcat(process, nodes)

    return result, {}
